"""
Convolutional Transformer

Largely inspired by https://github.com/lucidrains/vit-pytorch/blob/main/vit_pytorch/cvt.py

Original paper:
@article{wu2021cvt,
  title={Cvt: Introducing convolutions to vision transformers},
  author={Wu, Haiping and Xiao, Bin and Codella, Noel and Liu, Mengchen and Dai, Xiyang and Yuan, Lu and Zhang, Lei},
  journal={arXiv preprint arXiv:2103.15808},
  year={2021}
}
"""

import torch
import torch.nn as nn
import torch.nn.functional as F


def group_dict_by_key(cond, d):
    return_val = [dict(), dict()]
    for key in d.keys():
        match = bool(cond(key))
        ind = int(not match)
        return_val[ind][key] = d[key]
    return (*return_val,)


def group_by_key_prefix_and_remove_prefix(prefix, d):
    kwargs_with_prefix, kwargs = group_dict_by_key(lambda x: x.startswith(prefix), d)
    kwargs_without_prefix = dict(
        map(lambda x: (x[0][len(prefix) :], x[1]), tuple(kwargs_with_prefix.items()))
    )
    return kwargs_without_prefix, kwargs


# classes


class LayerNorm(nn.Module):
    """Layer normalization, but done in channel dimension #1"""

    def __init__(self, dim, eps=1e-5):
        super().__init__()
        self.eps = eps
        self.g = nn.Parameter(torch.ones(1, dim, 1, 1))
        self.b = nn.Parameter(torch.zeros(1, dim, 1, 1))

    def forward(self, x):
        std = torch.var(x, dim=1, unbiased=False, keepdim=True).sqrt()
        mean = torch.mean(x, dim=1, keepdim=True)
        return (x - mean) / (std + self.eps) * self.g + self.b


class PreNorm(nn.Module):
    """Pre-Normalization layer"""

    def __init__(self, dim, fn):
        super().__init__()
        self.norm = LayerNorm(dim)
        self.fn = fn

    def forward(self, x, **kwargs):
        x = self.norm(x)
        return self.fn(x, **kwargs)


class FeedForward(nn.Module):
    """Convolutional projection in the transformer."""

    def __init__(self, dim, mult=4, dropout=0.0):
        super().__init__()
        self.net = nn.Sequential(
            nn.Conv2d(dim, dim * mult, 1),  # 1x1 convolution
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Conv2d(dim * mult, dim, 1),  # 1x1 convolution
            nn.Dropout(dropout),
        )

    def forward(self, x):
        return self.net(x)


class DepthWiseConv2d(nn.Module):
    """Depthwise convolutional layer"""

    def __init__(self, dim_in, dim_out, kernel_size, padding, stride, bias=True):
        super().__init__()
        self.net = nn.Sequential(
            nn.Conv2d(
                dim_in,
                dim_in,
                kernel_size=kernel_size,
                padding=padding,
                groups=dim_in,
                stride=stride,
                bias=bias,
            ),
            nn.BatchNorm2d(dim_in),
            nn.Conv2d(dim_in, dim_out, kernel_size=1, bias=bias),
        )

    def forward(self, x):
        return self.net(x)


class Attention(nn.Module):
    """Custom Attention layer"""

    def __init__(
        self, dim, proj_kernel, kv_proj_stride, heads=8, dim_head=64, dropout=0.0
    ):
        super().__init__()
        inner_dim = dim_head * heads
        padding = proj_kernel // 2
        self.heads = heads
        self.scale = dim_head**-0.5

        self.attend = nn.Softmax(dim=-1)

        self.to_q = DepthWiseConv2d(
            dim, inner_dim, proj_kernel, padding=padding, stride=1, bias=False
        )
        self.to_kv = DepthWiseConv2d(
            dim,
            inner_dim * 2,
            proj_kernel,
            padding=padding,
            stride=kv_proj_stride,
            bias=False,
        )

        self.to_out = nn.Sequential(nn.Conv2d(inner_dim, dim, 1), nn.Dropout(dropout))

    def forward(self, x):
        from einops import rearrange, repeat
        from einops.layers.torch import Rearrange

        shape = x.shape
        b, n, _, y, h = *shape, self.heads
        q, k, v = (self.to_q(x), *self.to_kv(x).chunk(2, dim=1))
        q, k, v = map(
            lambda t: rearrange(t, "b (h d) x y -> (b h) (x y) d", h=h), (q, k, v)
        )

        dots = torch.einsum("b i d, b j d -> b i j", q, k) * self.scale

        attn = self.attend(dots)

        out = torch.einsum("b i j, b j d -> b i d", attn, v)
        out = rearrange(out, "(b h) (x y) d -> b (h d) x y", h=h, y=y)
        return self.to_out(out)


class Transformer(nn.Module):
    """Custom Transformer layer."""

    def __init__(
        self,
        dim,
        proj_kernel,
        kv_proj_stride,
        depth,
        heads,
        dim_head=64,
        mlp_mult=4,
        dropout=0.0,
    ):
        super().__init__()
        self.layers = nn.ModuleList([])
        for _ in range(depth):
            self.layers.append(
                nn.ModuleList(
                    [
                        PreNorm(
                            dim,
                            Attention(
                                dim,
                                proj_kernel=proj_kernel,
                                kv_proj_stride=kv_proj_stride,
                                heads=heads,
                                dim_head=dim_head,
                                dropout=dropout,
                            ),
                        ),
                        PreNorm(dim, FeedForward(dim, mlp_mult, dropout=dropout)),
                    ]
                )
            )

    def forward(self, x):
        for attn, ff in self.layers:
            x = attn(x) + x
            x = ff(x) + x
        return x


class CvT(nn.Module):
    """Convolutional Transformer module.

    Adapted for self-supervised training

    Attributes
    ----------
    s{i}_emb_dim: int
        Embedding dimention at stage i

    s{i}_emb_kernel: int
        Convolutional kernel size at stage i

    s{i}_emb_stride: int
        Convolutional stride at stage i

    s{i}_kv_proj_stride: int
        Convolutional stride in the convolutional projection layers at stage i

    s{i}_heads: int
        Number of attention heads at stage i

    s{i}_depth: int
        Transformer depth at stage i

    s{i}_mlp_mult: int
        MLP ratio at stage i

    dropout: float
        Dropout ratio
    """

    # sample rate and embedding sizes are required model attributes for the HEAR API
    sample_rate = 16000
    embedding_size = 2048
    scene_embedding_size = embedding_size
    timestamp_embedding_size = embedding_size

    def __init__(
        self,
        *,
        s1_emb_dim=64,
        s1_emb_kernel=7,
        s1_emb_stride=4,
        s1_proj_kernel=3,
        s1_kv_proj_stride=2,
        s1_heads=1,
        s1_depth=1,
        s1_mlp_mult=4,
        s2_emb_dim=192,
        s2_emb_kernel=3,
        s2_emb_stride=2,
        s2_proj_kernel=3,
        s2_kv_proj_stride=2,
        s2_heads=3,
        s2_depth=2,
        s2_mlp_mult=4,
        s3_emb_dim=384,
        s3_emb_kernel=3,
        s3_emb_stride=2,
        s3_proj_kernel=3,
        s3_kv_proj_stride=2,
        s3_heads=6,
        s3_depth=10,
        s3_mlp_mult=4,
        dropout=0.0,
        pool="mean",
    ):
        super().__init__()
        kwargs = dict(locals())

        dim = 1
        layers = []

        for prefix in ("s1", "s2", "s3"):
            config, kwargs = group_by_key_prefix_and_remove_prefix(f"{prefix}_", kwargs)

            layers.append(
                nn.Sequential(
                    nn.Conv2d(
                        dim,
                        config["emb_dim"],
                        kernel_size=config["emb_kernel"],
                        padding=(config["emb_kernel"] // 2),
                        stride=config["emb_stride"],
                    ),
                    LayerNorm(config["emb_dim"]),
                    Transformer(
                        dim=config["emb_dim"],
                        proj_kernel=config["proj_kernel"],
                        kv_proj_stride=config["kv_proj_stride"],
                        depth=config["depth"],
                        heads=config["heads"],
                        mlp_mult=config["mlp_mult"],
                        dropout=dropout,
                    ),
                )
            )

            dim = config["emb_dim"]

        self.pool = pool
        assert self.pool in ["mean", "max", "mean+max"]

        if self.pool == "mean":
            self.pool_layers = nn.Sequential(nn.AdaptiveAvgPool2d(1), nn.Flatten())
        elif self.pool == "max":
            self.pool_layers = (nn.Sequential(nn.AdaptiveMaxPool2d(1), nn.Flatten()),)
        else:
            self.pool_layers = nn.Sequential(nn.Identity())

        self.layers = nn.Sequential(
            *layers,
            *self.pool_layers,
        )

    def forward(self, x):
        x = self.layers(x)

        if self.pool == "mean+max":
            x = x.permute(0, 3, 2, 1)  # (batch, time, mel, ch)
            B, T, D, C = x.shape
            x = x.reshape((B, T, C * D))  # (batch, time, mel*ch)
            x = x.mean(1) + x.amax(1)
        return x



import torch

from .models.passt import get_model as get_model_passt
from .models.preprocess import AugmentMelSTFT
from .wrapper import PasstBasicWrapper


def load_model(model_path="", mode="all", **kwds):
    model = get_basic_model(mode=mode, **kwds)
    return model


def get_scene_embeddings(audio, model):
    """
    audio: n_sounds x n_samples of mono audio in the range [-1, 1]. All sounds in a batch will be padded/trimmed to the same length.
    model: Loaded Model.
    Returns:
    embedding: A float32 Tensor with shape (n_sounds, model.scene_embedding_size).
    """
    return model.get_scene_embeddings(audio)


def get_timestamp_embeddings(audio, model):
    """
    audio: n_sounds x n_samples of mono audio in the range [-1, 1]. All sounds in a batch will be padded/trimmed to the same length.
    model: Loaded Model.
    Returns:
    embedding: A float32 Tensor with shape (n_sounds, n_timestamps, model.timestamp_embedding_size).
    timestamps: A float32 Tensor with shape (`n_sounds, n_timestamps). Centered timestamps in milliseconds corresponding to each embedding in the output.
    """
    return get_basic_timestamp_embeddings(audio, model)


def get_basic_timestamp_embeddings(audio, model):
    """
    audio: n_sounds x n_samples of mono audio in the range [-1, 1]. All sounds in a batch will be padded/trimmed to the same length.
    model: Loaded Model.
    Returns:
    embedding: A float32 Tensor with shape (n_sounds, n_timestamps, model.timestamp_embedding_size).
    timestamps: A float32 Tensor with shape (`n_sounds, n_timestamps). Centered timestamps in milliseconds corresponding to each embedding in the output.
    """
    return model.get_timestamp_embeddings(audio)


def get_basic_model(**kwargs):
    mel = AugmentMelSTFT(
        n_mels=128,
        sr=32000,
        win_length=800,
        hopsize=320,
        n_fft=1024,
        freqm=48,
        timem=192,
        htk=False,
        fmin=0.0,
        fmax=None,
        norm=1,
        fmin_aug_range=10,
        fmax_aug_range=2000,
    )

    net = get_model_passt(arch="passt_s_swa_p16_128_ap476")
    model = PasstBasicWrapper(mel=mel, net=net, **kwargs)
    return model



"""
The setting of Superb SID

Authors
  * Po-Han Chi 2021
  * Leo 2022
"""

import logging
import math
import pickle
from dataclasses import dataclass
from pathlib import Path
from typing import List

import pandas as pd
from omegaconf import MISSING

from s3prl.dataio.corpus.voxceleb1sid import VoxCeleb1SID
from s3prl.dataio.dataset import EncodeCategory, LoadAudio
from s3prl.dataio.encoder.category import CategoryEncoder
from s3prl.dataio.sampler import FixedBatchSizeBatchSampler
from s3prl.nn.linear import MeanPoolingLinear

from .run import Common

logger = logging.getLogger(__name__)


__all__ = [
    "voxceleb1_for_sid",
    "SuperbSID",
]


def voxceleb1_for_sid(
    target_dir: str,
    cache_dir: str,
    dataset_root: str,
    n_jobs: int = 6,
    get_path_only: bool = False,
):
    """
    Prepare VoxCeleb1 for SID following :obj:`SuperbSID.prepare_data` format.

    Args:
        dataset_root (str): The root path of VoxCeleb1
        n_jobs (int): to speed up the corpus parsing procedure
        **others: refer to :obj:`SuperbSID.prepare_data`
    """
    target_dir = Path(target_dir)

    train_path = target_dir / f"train.csv"
    valid_path = target_dir / f"valid.csv"
    test_paths = [target_dir / f"test.csv"]

    if get_path_only:
        return train_path, valid_path, test_paths

    corpus = VoxCeleb1SID(dataset_root, n_jobs)
    train_data, valid_data, test_data = corpus.data_split

    def dict_to_csv(data_dict, csv_path):
        keys = sorted(list(data_dict.keys()))
        fields = sorted(data_dict[keys[0]].keys())
        data = dict()
        for field in fields:
            data[field] = []
            for key in keys:
                data[field].append(data_dict[key][field])
        data["id"] = keys
        df = pd.DataFrame(data)
        df.to_csv(csv_path, index=False)

    dict_to_csv(train_data, train_path)
    dict_to_csv(valid_data, valid_path)
    dict_to_csv(test_data, test_paths[0])

    return train_path, valid_path, test_paths


class SuperbSID(Common):
    """
    The standard SUPERB SID task
    """

    def default_config(self) -> dict:
        return dict(
            start=0,
            stop=None,
            target_dir=MISSING,
            cache_dir=None,
            remove_all_cache=False,
            prepare_data=dict(
                dataset_root=MISSING,
            ),
            build_encoder=dict(),
            build_dataset=dict(
                train=dict(
                    max_secs=8.0,
                ),
            ),
            build_batch_sampler=dict(
                train=dict(
                    batch_size=8,
                    shuffle=True,
                ),
                valid=dict(
                    batch_size=1,
                ),
                test=dict(
                    batch_size=1,
                ),
            ),
            build_upstream=dict(
                name=MISSING,
            ),
            build_featurizer=dict(
                layer_selections=None,
                normalize=False,
            ),
            build_downstream=dict(
                hidden_size=256,
            ),
            build_model=dict(
                upstream_trainable=False,
            ),
            build_task=dict(),
            build_optimizer=dict(
                name="Adam",
                conf=dict(
                    lr=1.0e-4,
                ),
            ),
            build_scheduler=dict(
                name="ExponentialLR",
                gamma=0.9,
            ),
            save_model=dict(),
            save_task=dict(),
            train=dict(
                total_steps=200000,
                log_step=500,
                eval_step=5000,
                save_step=1000,
                gradient_clipping=1.0,
                gradient_accumulate=4,
                valid_metric="accuracy",
                valid_higher_better=True,
                auto_resume=True,
                resume_ckpt_dir=None,
            ),
            evaluate=dict(),
        )

    def prepare_data(
        self,
        prepare_data: dict,
        target_dir: str,
        cache_dir: str,
        get_path_only: bool = False,
    ):
        """
        Prepare the task-specific data metadata (path, labels...).
        By default call :obj:`voxceleb1_for_sid` with :code:`**prepare_data`

        Args:
            prepare_data (dict): same in :obj:`default_config`, support arguments in :obj:`voxceleb1_for_sid`
            target_dir (str): Parse your corpus and save the csv file into this directory
            cache_dir (str): If the parsing or preprocessing takes too long time, you can save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            get_path_only (str): Directly return the filepaths no matter they exist or not.

        Returns:
            tuple

            1. train_path (str)
            2. valid_path (str)
            3. test_paths (List[str])

            Each path (str) should be a csv file containing the following columns:

            ====================  ====================
            column                description
            ====================  ====================
            id                    (str) - the unique id for this data point
            wav_path              (str) - the absolute path of the waveform file
            label                 (str) - a string label of the waveform
            start_sec             (float) - optional, load the waveform from :code:`start_sec` seconds. If not presented or is :code:`math.nan`, load from the beginning.
            end_sec               (float) - optional, load the waveform from :code:`end_sec` seconds. If not presented or is :code:`math.nan`, load to the end.
            ====================  ====================
        """
        return voxceleb1_for_sid(
            **self._get_current_arguments(flatten_dict="prepare_data")
        )

    def build_encoder(
        self,
        build_encoder: dict,
        target_dir: str,
        cache_dir: str,
        train_csv_path: str,
        valid_csv_path: str,
        test_csv_paths: list,
        get_path_only: bool = False,
    ):
        """
        Build the encoder (for the labels) given the data metadata, and return the saved encoder path.
        By default generate and save a :obj:`s3prl.dataio.encoder.CategoryEncoder` from the :code:`label` column of all the csv files.

        Args:
            build_encoder (dict): same in :obj:`default_config`, no argument supported for now
            target_dir (str): Save your encoder into this directory
            cache_dir (str): If the preprocessing takes too long time, you can save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            train_csv_path (str): the train path from :obj:`prepare_data`
            valid_csv_path (str): the valid path from :obj:`prepare_data`
            test_csv_paths (List[str]): the test paths from :obj:`prepare_data`
            get_path_only (str): Directly return the filepaths no matter they exist or not.

        Returns:
            str

            encoder_path: The encoder should be saved in the pickle format
        """
        encoder_path = Path(target_dir) / "encoder.pkl"
        if get_path_only:
            return encoder_path

        train_csv = pd.read_csv(train_csv_path)
        valid_csv = pd.read_csv(valid_csv_path)
        test_csvs = [pd.read_csv(path) for path in test_csv_paths]
        all_csv = pd.concat([train_csv, valid_csv, *test_csvs])

        labels = all_csv["label"].tolist()
        encoder = CategoryEncoder(labels)
        with open(encoder_path, "wb") as f:
            pickle.dump(encoder, f)

        return encoder_path

    def build_dataset(
        self,
        build_dataset: dict,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        encoder_path: str,
        frame_shift: int,
    ):
        """
        Build the dataset for train/valid/test.

        Args:
            build_dataset (dict): same in :obj:`default_config`. with :code:`train`, :code:`valid`, :code:`test` keys, each
                is a dictionary with the following supported options:

                ====================  ====================
                key                   description
                ====================  ====================
                max_secs              (float) - If a waveform is longer than :code:`max_secs` seconds, randomly crop the waveform into :code:`max_secs` seconds
                sox_effects           (List[List[str]]) - If not None, apply sox effects on the utterance
                ====================  ====================

            target_dir (str): Current experiment directory
            cache_dir (str): If the preprocessing takes too long time, you can save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            mode (str): train/valid/test
            data_csv (str): The metadata csv file for the specific :code:`mode`
            encoder_path (str): The pickled encoder path for encoding the labels

        Returns:
            torch Dataset

            For all train/valid/test mode, the dataset should return each item as a dictionary
            containing the following keys:

            ====================  ====================
            key                   description
            ====================  ====================
            x                     (torch.FloatTensor) - the waveform in (seq_len, 1)
            x_len                 (int) - the waveform length :code:`seq_len`
            class_id              (int) - the encoded class id
            label                 (str) - the class name
            unique_name           (str) - the unique id for this datapoint
            ====================  ====================
        """

        @dataclass
        class Config:
            train: dict = None
            valid: dict = None
            test: dict = None

        conf = Config(**build_dataset)

        assert mode in ["train", "valid", "test"]
        if mode == "train":
            conf = conf.train or {}
        elif mode == "valid":
            conf = conf.valid or {}
        elif mode == "test":
            conf = conf.test or {}

        @dataclass
        class SplitConfig:
            max_secs: float = None
            sox_effects: List[List[str]] = None

        conf = SplitConfig(**conf)

        csv = pd.read_csv(data_csv)

        start_secs = None
        if "start_sec" in csv.columns:
            start_secs = csv["start_sec"].tolist()
            start_secs = [None if math.isnan(sec) else sec for sec in start_secs]

        end_secs = None
        if "end_sec" in csv.columns:
            end_secs = csv["end_sec"].tolist()
            end_secs = [None if math.isnan(sec) else sec for sec in end_secs]

        audio_loader = LoadAudio(
            csv["wav_path"].tolist(),
            start_secs,
            end_secs,
            max_secs=conf.max_secs,
            sox_effects=conf.sox_effects,
        )

        with open(encoder_path, "rb") as f:
            encoder = pickle.load(f)

        label_encoder = EncodeCategory(csv["label"].tolist(), encoder)
        ids = csv["id"].tolist()

        class Dataset:
            def __len__(self):
                return len(ids)

            def __getitem__(self, index: int):
                audio = audio_loader[index]
                label = label_encoder[index]
                return {
                    "x": audio["wav"],
                    "x_len": audio["wav_len"],
                    "label": label["label"],
                    "class_id": label["class_id"],
                    "unique_name": ids[index],
                }

        dataset = Dataset()
        return dataset

    def build_batch_sampler(
        self,
        build_batch_sampler: dict,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        dataset,
    ):
        """
        Return the batch sampler for torch DataLoader.

        Args:
            build_batch_sampler (dict): same in :obj:`default_config`

                ====================  ====================
                key                   description
                ====================  ====================
                train                 (dict) - arguments for :obj:`FixedBatchSizeBatchSampler`
                valid                 (dict) - arguments for :obj:`FixedBatchSizeBatchSampler`
                test                  (dict) - arguments for :obj:`FixedBatchSizeBatchSampler`
                ====================  ====================

            target_dir (str): Current experiment directory
            cache_dir (str): If the preprocessing takes too long time, save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            mode (str): train/valid/test
            data_csv (str): the :code:`mode` specific csv from :obj:`prepare_data`
            dataset: the dataset from :obj:`build_dataset`

        Returns:
            batch sampler for torch DataLoader
        """

        def _superb_sid_batch_sampler(
            train: dict = None, valid: dict = None, test: dict = None
        ):
            train = train or {}
            valid = valid or {}
            test = test or {}

            if mode == "train":
                sampler = FixedBatchSizeBatchSampler(dataset, **train)
            elif mode == "valid":
                sampler = FixedBatchSizeBatchSampler(dataset, **valid)
            elif mode == "test":
                sampler = FixedBatchSizeBatchSampler(dataset, **test)

            return sampler

        return _superb_sid_batch_sampler(**build_batch_sampler)

    def build_downstream(
        self,
        build_downstream: dict,
        downstream_input_size: int,
        downstream_output_size: int,
        downstream_input_stride: int,
    ):
        """
        Return the task-specific downstream model.
        By default build the :obj:`MeanPoolingLinear` model

        Args:
            build_downstream (dict): same in :obj:`default_config`, support arguments of :obj:`MeanPoolingLinear`
            downstream_input_size (int): the required input size of the model
            downstream_output_size (int): the required output size of the model
            downstream_input_stride (int): the input feature's stride (from 16 KHz)

        Returns:
            :obj:`s3prl.nn.interface.AbsUtteranceModel`
        """
        model = MeanPoolingLinear(
            downstream_input_size, downstream_output_size, **build_downstream
        )
        return model



from .expert import UpstreamExpert as _UpstreamExpert


def passt_base(**kwds):
    return _UpstreamExpert("base", **kwds)


def passt_base2level(**kwds):
    return _UpstreamExpert("base2level", **kwds)


def passt_base2levelmel(**kwds):
    return _UpstreamExpert("base2levelmel", **kwds)


def passt_base20sec(**kwds):
    return _UpstreamExpert("base20sec", **kwds)


def passt_base30sec(**kwds):
    return _UpstreamExpert("base30sec", **kwds)


def passt_hop100base(**kwds):
    return _UpstreamExpert("hop100base", **kwds)


def passt_hop100base2lvl(**kwds):
    return _UpstreamExpert("hop100base2lvl", **kwds)


def passt_hop100base2lvlmel(**kwds):
    return _UpstreamExpert("hop100base2lvlmel", **kwds)


def passt_hop160base(**kwds):
    return _UpstreamExpert("hop160base", **kwds)


def passt_hop160base2lvl(**kwds):
    return _UpstreamExpert("hop160base2lvl", **kwds)


def passt_hop160base2lvlmel(**kwds):
    return _UpstreamExpert("hop160base2lvlmel", **kwds)


# FIXME: url seems corrupted
# def passt_openmic2008(**kwds):
#     return _UpstreamExpert("openmic2008", **kwds)



# Copyright Hear Benchmark Team
# Copyright Shu-wen Yang (refactor from https://github.com/hearbenchmark/hear-eval-kit)

from typing import List

import torch

from s3prl.dataio.encoder.category import CategoryEncoder
from s3prl.task.base import Task

from ._hear_score import available_scores, validate_score_return_type

__all__ = ["ScenePredictionTask"]


class OneHotToCrossEntropyLoss(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.loss = torch.nn.CrossEntropyLoss()

    def forward(self, y_hat: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        assert torch.all(torch.sum(y, dim=1) == y.new_ones(y.shape[0]))
        y = y.argmax(dim=1)
        return self.loss(y_hat, y)


class ScenePredictionTask(Task):
    def __init__(
        self,
        model: torch.nn.Module,
        category: CategoryEncoder,
        prediction_type: str,
        scores: List[str],
    ):
        super().__init__()
        self.model = model

        self.label_to_idx = {
            str(category.decode(idx)): idx for idx in range(len(category))
        }
        self.idx_to_label = {
            idx: str(category.decode(idx)) for idx in range(len(category))
        }
        self.scores = [
            available_scores[score](label_to_idx=self.label_to_idx) for score in scores
        ]

        if prediction_type == "multilabel":
            self.activation: torch.nn.Module = torch.nn.Sigmoid()
            self.logit_loss = torch.nn.BCEWithLogitsLoss()
        elif prediction_type == "multiclass":
            self.activation = torch.nn.Softmax(dim=-1)
            self.logit_loss = OneHotToCrossEntropyLoss()
        else:
            raise ValueError(f"Unknown prediction_type {prediction_type}")

    def predict(self, x, x_len):
        logits, _ = self.model(x, x_len)
        prediction = self.activation(logits)
        return prediction, logits

    def forward(
        self, _mode: str, x, x_len, y, labels, unique_name: str, _dump_dir: str = None
    ):
        y_pr, y_hat = self.predict(x, x_len)
        loss = self.logit_loss(y_hat.float(), y.float())

        cacheable = dict(
            loss=loss.detach().cpu().item(),
            label=y.detach().cpu().unbind(dim=0),  # (batch_size, num_class)
            logit=y_hat.detach().cpu().unbind(dim=0),  # (batch_size, num_class)
            prediction=y_pr.detach().cpu().unbind(dim=0),  # (batch_size, num_class)
        )

        return loss, cacheable

    def log_scores(self, score_args):
        """Logs the metric score value for each score defined for the model"""
        assert hasattr(self, "scores"), "Scores for the model should be defined"
        end_scores = {}
        # The first score in the first `self.scores` is the optimization criterion
        for score in self.scores:
            score_ret = score(*score_args)
            validate_score_return_type(score_ret)
            # If the returned score is a tuple, store each subscore as separate entry
            if isinstance(score_ret, tuple):
                end_scores[f"{score}"] = score_ret[0][1]
                # All other scores will also be logged
                for subscore, value in score_ret:
                    end_scores[f"{score}_{subscore}"] = value
            elif isinstance(score_ret, float):
                end_scores[f"{score}"] = score_ret
            else:
                raise ValueError(
                    f"Return type {type(score_ret)} is unexpected. Return type of "
                    "the score function should either be a "
                    "tuple(tuple) or float."
                )
        return end_scores

    def reduction(
        self,
        _mode: str,
        cached_results: List[dict],
        _dump_dir: str = None,
    ):
        result = self.parse_cached_results(cached_results)

        target = torch.stack(result["label"], dim=0)
        prediction_logit = torch.stack(result["logit"], dim=0)
        prediction = torch.stack(result["prediction"], dim=0)

        loss = self.logit_loss(prediction_logit, target)

        logs = dict(
            loss=loss.detach().cpu().item(),
        )

        if _mode in ["valid", "test"]:
            logs.update(
                self.log_scores(
                    score_args=(
                        prediction.detach().cpu().numpy(),
                        target.detach().cpu().numpy(),
                    ),
                )
            )

        return logs



# --------------------------------------------------------
# LightHuBERT: Lightweight and Configurable Speech Representation Learning with Once-for-All Hidden-Unit BERT (https://arxiv.org/pdf/2203.15610.pdf)
# Github source: https://github.com/mechanicalsea/lighthubert
# Licensed under The MIT License [see LICENSE for details]
# Based on fairseq code bases
# https://github.com/pytorch/fairseq
# --------------------------------------------------------

import math
from typing import Dict, Optional, Tuple

import torch
import torch.nn as nn
from torch import Tensor

from ..functional.sliding_attn import (
    global_attention_forward,
    slide_window_attention_forward,
)
from .fairseq_modules import quant_noise
from .scaling_linear import SLinear


class SMHA(nn.Module):
    """SMHA (Scaling MultiheadAttention): variable input (i.e., output) size and heads number.
    where in_embed_dim = out_embed_dim, qkv_embed_dim = 64 * num_heads

    wav2vec2:MultiheadAttention(
        embed_dim,
        num_heads,
        dropout=...,
        self_attention=True,
    )
    Module: self.k_proj, self.v_proj, self.q_proj, self.out_proj
    None: self.bias_k, self.bias_v

    __base__: fairseq.modules.MultiheadAttention
    """

    def __init__(
        self,
        embed_dim,
        num_heads,
        kdim=None,
        vdim=None,
        dropout=0.0,
        bias=True,
        add_bias_kv=False,
        add_zero_attn=False,
        self_attention=False,
        encoder_decoder_attention=False,
        q_noise=0.0,  # if 0, no quant_noise
        qn_block_size=8,
        sliding_attn_window="global",
        slide_mode="stride",
    ):
        super().__init__()
        self.embed_dim = embed_dim
        self.kdim = kdim if kdim is not None else embed_dim
        self.vdim = vdim if vdim is not None else embed_dim
        self.qkv_same_dim = self.kdim == embed_dim and self.vdim == embed_dim

        self.num_heads = num_heads
        self.dropout_module = nn.Dropout(dropout)

        self.head_dim = embed_dim // num_heads
        assert (
            self.head_dim * num_heads == self.embed_dim
        ), "embed_dim must be divisible by num_heads"
        self.scaling = self.head_dim**-0.5

        self.self_attention = self_attention
        self.encoder_decoder_attention = encoder_decoder_attention

        assert not self.self_attention or self.qkv_same_dim, (
            "Self-attention requires query, key and " "value to be of the same size"
        )

        self.k_proj = quant_noise(
            SLinear(self.kdim, embed_dim, bias=bias), q_noise, qn_block_size
        )  # lighthubert component
        self.v_proj = quant_noise(
            SLinear(self.vdim, embed_dim, bias=bias), q_noise, qn_block_size
        )  # lighthubert component
        self.q_proj = quant_noise(
            SLinear(embed_dim, embed_dim, bias=bias), q_noise, qn_block_size
        )  # lighthubert component

        self.out_proj = quant_noise(
            SLinear(embed_dim, embed_dim, bias=bias), q_noise, qn_block_size
        )  # lighthubert component

        if add_bias_kv:
            self.bias_k = nn.Parameter(torch.Tensor(1, 1, embed_dim))
            self.bias_v = nn.Parameter(torch.Tensor(1, 1, embed_dim))
        else:
            self.bias_k = self.bias_v = None

        self.add_zero_attn = add_zero_attn

        self.reset_parameters()

        self.onnx_trace = False

        # scaling attention
        self.slide_mode = slide_mode
        self.sliding_attn_window = sliding_attn_window

        self.sample_qkv_embed_dim = None
        self.sample_num_heads = None
        self.sample_in_embed_dim = None
        self.sample_attn_swz = "global"

    def prepare_for_onnx_export_(self):
        self.onnx_trace = True

    def reset_parameters(self):
        if self.qkv_same_dim:
            # Empirically observed the convergence to be much better with
            # the scaled initialization
            nn.init.xavier_uniform_(self.k_proj.weight, gain=1 / math.sqrt(2))
            nn.init.xavier_uniform_(self.v_proj.weight, gain=1 / math.sqrt(2))
            nn.init.xavier_uniform_(self.q_proj.weight, gain=1 / math.sqrt(2))
        else:
            nn.init.xavier_uniform_(self.k_proj.weight)
            nn.init.xavier_uniform_(self.v_proj.weight)
            nn.init.xavier_uniform_(self.q_proj.weight)

        nn.init.xavier_uniform_(self.out_proj.weight)
        if self.out_proj.bias is not None:
            nn.init.constant_(self.out_proj.bias, 0.0)
        if self.bias_k is not None:
            nn.init.xavier_normal_(self.bias_k)
        if self.bias_v is not None:
            nn.init.xavier_normal_(self.bias_v)

    def set_slide_mode(self, slide_mode):
        assert slide_mode in ["stride", "mask"]
        self.slide_mode = slide_mode

    def set_sample_config(
        self,
        sample_qkv_embed_dim: int,
        sample_num_heads: int,
        sample_in_embed_dim: int,
        sample_attn_swz="global",
    ):
        if sample_qkv_embed_dim is None:
            sample_qkv_embed_dim = self.embed_dim
        if sample_num_heads is None:
            sample_num_heads = self.num_heads
        if sample_in_embed_dim is None:
            sample_in_embed_dim = self.embed_dim
        if sample_attn_swz is None or sample_attn_swz == "global":
            sample_attn_swz = "global"
        assert sample_num_heads * 64 == sample_qkv_embed_dim, ValueError(
            f"heads num {sample_num_heads} * 64 != qkv dim {sample_qkv_embed_dim}"
        )
        self.sample_qkv_embed_dim = sample_qkv_embed_dim
        self.sample_num_heads = sample_num_heads
        self.sample_in_embed_dim = sample_in_embed_dim
        self.sample_attn_swz = sample_attn_swz
        self._sample_parameters()

    def _sample_parameters(self):
        self.k_proj.set_sample_config(
            self.sample_in_embed_dim, self.sample_qkv_embed_dim
        )
        self.v_proj.set_sample_config(
            self.sample_in_embed_dim, self.sample_qkv_embed_dim
        )
        self.q_proj.set_sample_config(
            self.sample_in_embed_dim, self.sample_qkv_embed_dim
        )
        self.out_proj.set_sample_config(
            self.sample_qkv_embed_dim, self.sample_in_embed_dim
        )

    def forward(
        self,
        query,
        key: Optional[Tensor],
        value: Optional[Tensor],
        key_padding_mask: Optional[Tensor] = None,
        incremental_state: Optional[Dict[str, Dict[str, Optional[Tensor]]]] = None,
        need_weights: bool = True,
        static_kv: bool = False,
        attn_mask: Optional[Tensor] = None,
        before_softmax: bool = False,
        need_head_weights: bool = False,
    ) -> Tuple[Tensor, Optional[Tensor]]:
        """Implement via fairseq multihead_attention with longformer's sliding attention window

        Args:
            query, key, value: (seq_len, batch_size, hidden_dim)

            key_padding_mask: if provided, specified padding elements in the key will
                be ignored by the attention. This is an binary mask. When the value is True,
                the corresponding value on the attention layer will be filled with -inf.
                binary ByteTensor of shape `(batch, seq_len)` where padding elements are indicated by ``1``.
            attn_mask: 2D or 3D mask that prevents attention to certain positions.
                A 2D mask will be broadcasted for all the batches while a 3D mask
                allows to specify a different mask for the entries of each batch.
                When the value is 1, the corresponding value on the attention
                layer will be added with -1e4 (float16) or -1e8 (float32) or -1e2 (float8).
                `attn_mask[tgt_i, src_j] = 1` means that when calculating the embedding for `tgt_i`,
                we exclude (mask out) `src_j`. This is useful for strided self-attention.
        """
        if need_head_weights:
            need_weights = True

        tgt_len, bsz, embed_dim = query.size()
        src_len = tgt_len
        if self.sample_in_embed_dim is None:
            assert (
                embed_dim == self.embed_dim
            ), f"query dim {embed_dim} != {self.embed_dim}"
        else:
            assert (
                embed_dim == self.sample_in_embed_dim
            ), f"query dim {embed_dim} != {self.sample_in_embed_dim}"
        if key is not None:
            src_len, key_bsz, _ = key.size()
            if not torch.jit.is_scripting():
                assert key_bsz == bsz
                assert value is not None
                assert src_len, bsz == value.shape[:2]

        # incremental_state is None
        assert incremental_state is None
        saved_state = None

        if self.self_attention:
            q = self.q_proj(query)
            k = self.k_proj(query)
            v = self.v_proj(query)
        elif self.encoder_decoder_attention:
            # encoder-decoder attention
            q = self.q_proj(query)
            if key is None:
                assert value is None
                k = v = None
            else:
                k = self.k_proj(key)
                v = self.v_proj(key)

        else:
            assert key is not None and value is not None
            q = self.q_proj(query)
            k = self.k_proj(key)
            v = self.v_proj(value)
        q *= self.scaling

        if self.bias_k is not None:
            assert self.bias_v is not None
            k = torch.cat([k, self.bias_k[:, :, : k.size(-1)].repeat(1, bsz, 1)])
            v = torch.cat([v, self.bias_v[:, :, : v.size(-1)].repeat(1, bsz, 1)])
            if attn_mask is not None:
                attn_mask = torch.cat(
                    [attn_mask, attn_mask.new_zeros(attn_mask.size(0), 1)], dim=1
                )
            if key_padding_mask is not None:
                key_padding_mask = torch.cat(
                    [
                        key_padding_mask,
                        key_padding_mask.new_zeros(key_padding_mask.size(0), 1),
                    ],
                    dim=1,
                )

        sample_num_heads = (
            self.num_heads if self.sample_num_heads is None else self.sample_num_heads
        )
        q = (
            q.contiguous()
            .view(tgt_len, bsz * sample_num_heads, self.head_dim)
            .transpose(0, 1)
        )
        if k is not None:
            k = (
                k.contiguous()
                .view(-1, bsz * sample_num_heads, self.head_dim)
                .transpose(0, 1)
            )
        if v is not None:
            v = (
                v.contiguous()
                .view(-1, bsz * sample_num_heads, self.head_dim)
                .transpose(0, 1)
            )

        assert k is not None
        assert k.size(1) == src_len

        # This is part of a workaround to get around fork/join parallelism
        # not supporting Optional types.
        if key_padding_mask is not None and key_padding_mask.dim() == 0:
            key_padding_mask = None

        if key_padding_mask is not None:
            assert key_padding_mask.size(0) == bsz
            assert key_padding_mask.size(1) == src_len

        if self.add_zero_attn:
            assert v is not None
            src_len += 1
            k = torch.cat([k, k.new_zeros((k.size(0), 1) + k.size()[2:])], dim=1)
            v = torch.cat([v, v.new_zeros((v.size(0), 1) + v.size()[2:])], dim=1)
            if attn_mask is not None:
                attn_mask = torch.cat(
                    [attn_mask, attn_mask.new_zeros(attn_mask.size(0), 1)], dim=1
                )
            if key_padding_mask is not None:
                key_padding_mask = torch.cat(
                    [
                        key_padding_mask,
                        torch.zeros(key_padding_mask.size(0), 1).type_as(
                            key_padding_mask
                        ),
                    ],
                    dim=1,
                )

        if self.sample_attn_swz == "global":
            attn, unnormalized_attn_weights = global_attention_forward(
                q,
                k,
                v,
                attn_mask=attn_mask,
                key_padding_mask=key_padding_mask,
                num_heads=sample_num_heads,
                dropout_p=self.dropout_module.p,
                training=self.training,
            )
        else:
            attn, unnormalized_attn_weights = slide_window_attention_forward(
                q,
                k,
                v,
                self.sample_attn_swz,
                attn_mask=attn_mask,
                key_padding_mask=key_padding_mask,
                num_heads=sample_num_heads,
                dropout_p=self.dropout_module.p,
                training=self.training,
                mode=self.slide_mode,
            )

        assert list(attn.size()) == [bsz * sample_num_heads, tgt_len, self.head_dim]

        attn_embed_dim = embed_dim
        if self.sample_qkv_embed_dim is not None:
            attn_embed_dim = self.sample_qkv_embed_dim
        if self.onnx_trace and attn.size(1) == 1:
            # when ONNX tracing a single decoder step (sequence length == 1)
            # the transpose is a no-op copy before view, thus unnecessary
            attn = attn.contiguous().view(tgt_len, bsz, attn_embed_dim)
        else:
            attn = attn.transpose(0, 1).contiguous().view(tgt_len, bsz, attn_embed_dim)
        attn = self.out_proj(attn)
        attn_weights: Optional[Tensor] = None
        if need_weights and isinstance(unnormalized_attn_weights, Tensor):
            wsz = src_len
            if self.slide_mode == "stride" and not self.sample_attn_swz == "global":
                wsz = self.sample_attn_swz + 1
            attn_weights = unnormalized_attn_weights.view(
                bsz, sample_num_heads, tgt_len, wsz
            )
        return attn, attn_weights

    def calc_sampled_param_num(self):
        total_params = 0
        total_params += self.k_proj.calc_sampled_param_num()
        total_params += self.v_proj.calc_sampled_param_num()
        total_params += self.q_proj.calc_sampled_param_num()
        total_params += self.out_proj.calc_sampled_param_num()
        return total_params

    def get_complexity(self, sequence_length):
        total_flops = 0
        total_flops += self.k_proj.get_complexity(sequence_length)
        total_flops += self.v_proj.get_complexity(sequence_length)
        total_flops += self.q_proj.get_complexity(sequence_length)
        total_flops += self.out_proj.get_complexity(sequence_length)
        # attn
        swa = self.sample_attn_swz
        if swa == "global" or sequence_length <= swa // 2 + 1:
            swa = sequence_length
        total_flops += sequence_length * swa * self.sample_qkv_embed_dim
        # x
        total_flops += sequence_length * swa * self.sample_qkv_embed_dim
        return total_flops


if __name__ == "__main__":
    m = SMHA(768, 12)
    m.set_sample_config(768, 12, 768, 32)
    x = torch.empty((2, 5, 768))
    m(x, x, x)



"""
Speaker Verification with Softmax-based loss

Authors
  * Po-Han Chi 2021
  * Haibin Wu 2022
"""

import logging
from typing import List, Tuple

import torch
import torch.nn.functional as F
from tqdm import tqdm

from s3prl.dataio.encoder.category import CategoryEncoder
from s3prl.metric import accuracy, compute_eer, compute_minDCF
from s3prl.nn import amsoftmax, softmax

from . import Task

logger = logging.getLogger(__name__)


__all__ = ["SpeakerClassifier", "SpeakerVerification"]


class SpeakerClassifier(torch.nn.Module):
    """
    Attributes:
        input_size: int
        output_size: int
    """

    def __init__(self, input_size=3, output_size=4):
        super().__init__()
        self._input_size = input_size
        self._output_size = output_size

    @property
    def input_size(self):
        return self._input_size

    @property
    def output_size(self):
        return self._output_size

    def forward(self, x, x_len):
        """
        Args:
            x (torch.Tensor): (batch_size, timestemps, input_size)
            x_len (torch.LongTensor): (batch_size, )

        Return:
            output (torch.Tensor): (batch_size, output_size)
        """
        assert x.size(-1) == self.input_size
        output = torch.randn(x.size(0), self.output_size)
        assert output


class SpeakerVerification(Task):
    """
    model.output_size should match len(categories)

    Args:
        model (SpeakerClassifier):
            actual model or a callable config for the model
        categories (dict[str]):
            each key in the Dictionary is the final prediction content in str.
            use categories[key] to encode as numeric label
        test_trials (List[Tuple[int, str, str]]):
            each tuple in the list consists of (label, enroll_utt, test_utt)
        loss_type (str): softmax or amsoftmax
        loss_conf (dict): arguments for the loss_type class
    """

    def __init__(
        self,
        model: SpeakerClassifier,
        category: CategoryEncoder,
        test_trials: List[Tuple[int, str, str]] = None,
        loss_type: str = "amsoftmax",
        loss_conf: dict = None,
    ):
        super().__init__()
        self.model = model
        self.category = category
        self.trials = test_trials

        if loss_type == "amsoftmax":
            loss_cls = amsoftmax
        elif loss_type == "softmax":
            loss_cls = softmax
        else:
            raise ValueError(f"Unsupported loss_type {loss_type}")

        self.loss: torch.nn.Module = loss_cls(
            input_size=self.model.output_size,
            output_size=len(self.category),
            **loss_conf,
        )
        assert self.loss.output_size == len(category)

    def get_state(self):
        return {
            "loss_state": self.loss.state_dict(),
        }

    def set_state(self, state: dict):
        self.loss.load_state_dict(state["loss_state"])

    def predict(self, x: torch.Tensor, x_len: torch.LongTensor):
        """
        Args:
            x (torch.Tensor): (batch_size, timestamps, input_size)
            x_len (torch.LongTensor): (batch_size, )

        Return:
            torch.Tensor

            (batch_size, output_size)
        """
        spk_embeddings = self.model(x, x_len)
        return spk_embeddings

    def train_step(
        self,
        x: torch.Tensor,
        x_len: torch.LongTensor,
        class_id: torch.LongTensor,
        unique_name: List[str],
        _dump_dir: str = None,
    ):
        spk_embeddings = self.predict(x, x_len)
        loss, logits = self.loss(spk_embeddings, class_id)
        prediction = [index for index in logits.argmax(dim=-1).detach().cpu().tolist()]

        cacheable = dict(
            loss=loss.detach().cpu().item(),
            class_id=class_id.detach().cpu().tolist(),
            prediction=prediction,
            unique_name=unique_name,
        )

        return loss, cacheable

    def train_reduction(self, cached_results: list, _dump_dir: str = None):
        results = self.parse_cached_results(cached_results)
        acc = accuracy(results["prediction"], results["class_id"])
        loss = torch.FloatTensor(results["loss"]).mean().item()

        return dict(
            loss=loss,
            accuracy=acc,
        )

    def test_step(
        self,
        x: torch.Tensor,
        x_len: torch.LongTensor,
        unique_name: List[str],
        _dump_dir: str,
    ):
        """
        Args:
            x (torch.Tensor): (batch_size, timestamps, input_size)
            x_len: torch.LongTensor
            unique_name (List[str])

        Return:
            unique_name (List[str])
            output (torch.Tensor):
                speaker embeddings corresponding to unique_name
        """
        spk_embeddings = self.predict(x, x_len)

        cacheable = dict(
            unique_name=unique_name.tolist(),
            spk_embedding=spk_embeddings.detach().cpu().unbind(dim=0),
        )
        return None, cacheable

    def test_reduction(self, cached_results: List[dict], _dump_dir: str):
        results = self.parse_cached_results(cached_results)
        embeddings = {}
        for name, emb in zip(results["unique_name"], results["spk_embedding"]):
            embeddings[name] = emb

        trials = self.trials
        scores = []
        labels = []
        for label, enroll, test in tqdm(trials, desc="Test Scoring", total=len(trials)):
            enroll_embd = embeddings[enroll]
            test_embd = embeddings[test]
            score = F.cosine_similarity(enroll_embd, test_embd, dim=0).item()
            scores.append(score)
            labels.append(label)

        EER, EERthreshold = compute_eer(labels, scores)

        minDCF, minDCFthreshold = compute_minDCF(labels, scores, p_target=0.01)

        return dict(
            EER=EER,
            EERthreshold=EERthreshold.item(),
            minDCF=minDCF,
            minDCF_threshold=minDCFthreshold,
        )



import logging
from dataclasses import dataclass

import numpy as np

from s3prl.dataio.encoder.category import CategoryEncoder

from .base import AugmentedDynamicItemDataset, DataPipe

logger = logging.getLogger(__name__)


@dataclass
class BuildMultiClassTagging(DataPipe):
    sample_rate: int = 16000
    feat_frame_shift: int = 160
    intra_or_inter: str = "intra"

    # input
    start_sec_name: str = "start_sec"
    end_sec_name: str = "end_sec"
    segments_name: str = "segments"

    # output
    tag_name: str = "multiclass_tag"
    tag_len_name: str = "tag_len"
    category_name: str = "tag_category"
    all_category_name: str = "all_tag_category"

    def build_label(
        self,
        segments: dict,
        start_sec: int,
        end_sec: int,
        all_tag_category: CategoryEncoder,
    ):
        frame_num = round(
            (end_sec - start_sec) * self.sample_rate / self.feat_frame_shift
        )

        if self.intra_or_inter == "inter":
            category = all_tag_category
        elif self.intra_or_inter == "intra":
            classes_in_this_segment = sorted(segments.keys())
            category = CategoryEncoder(list(classes_in_this_segment))
        else:
            raise ValueError("Only 'inter' or 'intra' is supported")

        T = np.zeros((frame_num, len(category)), dtype=np.int32)
        for class_name, start_ends in segments.items():
            class_idx = category.encode(class_name)
            for seg_start, seg_end in start_ends:
                rel_seg_start = rel_seg_end = None
                if start_sec <= seg_start and seg_start < end_sec:
                    rel_seg_start = seg_start - start_sec
                if start_sec < seg_end and seg_end <= end_sec:
                    rel_seg_end = seg_end - start_sec
                if rel_seg_start is not None or rel_seg_end is not None:
                    rel_seg_start_frame = (
                        round(rel_seg_start * self.sample_rate / self.feat_frame_shift)
                        if rel_seg_start is not None
                        else None
                    )
                    rel_seg_end_frame = (
                        round(rel_seg_end * self.sample_rate / self.feat_frame_shift)
                        if rel_seg_end is not None
                        else None
                    )
                    T[rel_seg_start_frame:rel_seg_end_frame, class_idx] = 1
        return T, frame_num, category

    def forward(
        self, dataset: AugmentedDynamicItemDataset
    ) -> AugmentedDynamicItemDataset:
        if not dataset.has_tool(self.all_category_name):
            logger.warning(
                f"The input dataset does not contain '{self.all_category_name}'. Generate it on-the-fly. "
            )
            with dataset.output_keys_as([self.segments_name]):
                all_classes = set()
                for item in dataset:
                    segments = item[self.segments_name]
                    for class_name in segments.keys():
                        all_classes.add(class_name)
                all_classes = sorted(all_classes)

            dataset.add_tool(self.all_category_name, CategoryEncoder(all_classes))

        dataset.add_dynamic_item(
            self.build_label,
            takes=[
                self.segments_name,
                self.start_sec_name,
                self.end_sec_name,
                self.all_category_name,
            ],
            provides=[
                self.tag_name,
                self.tag_len_name,
                self.category_name,
            ],
        )
        return dataset



import json
import logging
from collections import defaultdict
from pathlib import Path
from typing import List

from joblib import Parallel, delayed
from tqdm import tqdm

logger = logging.getLogger(__name__)


__all__ = [
    "get_info",
]


def get_info(dataset, names: List[str], cache_dir: str = None, n_jobs: int = 6):
    logger.info(
        f"Getting info from dataset {dataset.__class__.__qualname__}: {' '.join(names)}"
    )
    if isinstance(cache_dir, (str, Path)):
        logger.info(f"Using cached info in {cache_dir}")
        cache_dir: Path = Path(cache_dir)
        cache_dir.mkdir(parents=True, exist_ok=True)

    try:
        data = dataset.getinfo(0)
        for name in names:
            assert name in data
    except:
        fn = dataset.__getitem__
    else:
        fn = dataset.getinfo

    def _get(idx):
        if isinstance(cache_dir, (str, Path)):
            cache_path: Path = Path(cache_dir) / f"{idx}.json"
            if cache_path.is_file():
                with cache_path.open() as f:
                    cached = json.load(f)
                    all_presented = True
                    for name in names:
                        if name not in cached:
                            all_presented = False
                    if all_presented:
                        return cached

        data = fn(idx)

        info = {}
        for name in names:
            info[name] = data[name]

        if isinstance(cache_dir, (str, Path)):
            cache_path: Path = Path(cache_dir) / f"{idx}.json"
            with cache_path.open("w") as f:
                json.dump(info, f)

        return info

    infos = Parallel(n_jobs=n_jobs, backend="threading")(
        delayed(_get)(idx) for idx in tqdm(range(len(dataset)))
    )

    organized_info = defaultdict(list)
    for info in infos:
        for k, v in info.items():
            organized_info[k].append(v)

    output = []
    for name in names:
        output.append(organized_info[name])

    if len(output) == 1:
        return output[0]
    else:
        return output



from collections import Counter

import pytest
from dotenv import dotenv_values

from s3prl.dataio.corpus.speech_commands import SpeechCommandsV1


def _class_counter(data_dict):
    counter = Counter()
    for data_id, data in data_dict.items():
        counter.update([data["class_name"]])
    return counter


@pytest.mark.corpus
def test_speech_commands():
    env = dotenv_values()
    corpus = SpeechCommandsV1(env["GSC1"], env["GSC1_TEST"])
    all_data = corpus.all_data
    classes = set([value["class_name"] for key, value in all_data.items()])
    assert len(classes) == 12, f"{classes}"

    train, valid, test = corpus.data_split
    train_class_counter = _class_counter(train)
    valid_class_counter = _class_counter(valid)
    test_class_counter = _class_counter(test)

    # These pre-defined numbers are obtained with the old DownstreamExpert
    assert train_class_counter == Counter(
        {
            "_unknown_": 32550,
            "stop": 1885,
            "on": 1864,
            "go": 1861,
            "yes": 1860,
            "no": 1853,
            "right": 1852,
            "up": 1843,
            "down": 1842,
            "left": 1839,
            "off": 1839,
            "_silence_": 6,
        }
    )
    assert valid_class_counter == Counter(
        {
            "_unknown_": 4221,
            "stop": 246,
            "on": 257,
            "go": 260,
            "yes": 261,
            "no": 270,
            "right": 256,
            "up": 260,
            "down": 264,
            "left": 247,
            "off": 256,
            "_silence_": 6,
        }
    )
    assert test_class_counter == Counter(
        {
            "_unknown_": 257,
            "stop": 249,
            "on": 246,
            "go": 251,
            "yes": 256,
            "no": 252,
            "right": 259,
            "up": 272,
            "down": 253,
            "left": 267,
            "off": 262,
            "_silence_": 257,
        }
    )



import pytest
import torch

from s3prl.nn import BeamDecoder


@pytest.mark.extra_dependency
def test_beam_decoder():
    decoder = BeamDecoder()
    emissions = torch.randn((4, 100, 31))
    emissions = torch.log_softmax(emissions, dim=2)
    hyps = decoder.decode(emissions)



import pytest
import torch

from s3prl.nn.common import UtteranceLevel
from s3prl.nn.pooling import (
    AttentiveStatisticsPooling,
    MeanPooling,
    SelfAttentivePooling,
    TemporalStatisticsPooling,
)


@pytest.mark.parametrize(
    "pooling_type",
    [
        "MeanPooling",
        "TemporalStatisticsPooling",
        "AttentiveStatisticsPooling",
        "SelfAttentivePooling",
    ],
)
def test_utterance_level_with_pooling(pooling_type: str):
    model = UtteranceLevel(256, 64, [128], "ReLU", None, pooling_type, None)
    output = model(torch.randn(32, 100, 256), torch.arange(32) + 1)
    assert output.shape == (32, 64)



from s3prl.dataset.base import DataPipe, SequentialDataPipe
from s3prl.dataset.chunking import UnfoldChunkBySec
from s3prl.dataset.common_pipes import LoadAudio, SetOutputKeys
from s3prl.dataset.multiclass_tagging import BuildMultiClassTagging


class HearTimestampDatapipe(SequentialDataPipe):
    def __init__(
        self,
        sample_rate: int = 16000,
        feat_frame_shift: int = 160,
    ):
        super().__init__(
            UnfoldChunkBySec(
                min_chunk_secs=4.0,
                max_chunk_secs=4.0,
                step_secs=4.0,
            ),
            LoadAudio(audio_sample_rate=sample_rate),
            BuildMultiClassTagging(
                sample_rate=sample_rate,
                feat_frame_shift=feat_frame_shift,
                intra_or_inter="inter",
                all_category_name="category",
            ),
            SetOutputKeys(
                dict(
                    x="wav",
                    x_len="wav_len",
                    y="multiclass_tag",
                    y_len="tag_len",
                    record_id="unchunked_id",
                    chunk_id="chunk_index",
                )
            ),
        )



# -*- coding: utf-8 -*-
# @Time    : 7/16/21 3:12 AM
# @Author  : Yuan Gong
# @Affiliation  : Massachusetts Institute of Technology
# @Email   : yuangong@mit.edu
# @File    : ast_models.py

# the unified ast models for all pretraining/fine-tuning tasks.

import logging

import numpy as np
import timm
import torch
import torch.nn as nn
from timm.models.layers import to_2tuple, trunc_normal_

logger = logging.getLogger(__name__)


# override the timm package to relax the input shape constraint.
class PatchEmbed(nn.Module):
    """Image to Patch Embedding"""

    def __init__(self, img_size=224, patch_size=16, in_chans=3, embed_dim=768):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        num_patches = (img_size[1] // patch_size[1]) * (img_size[0] // patch_size[0])
        self.img_size = img_size
        self.patch_size = patch_size
        self.num_patches = num_patches

        self.proj = nn.Conv2d(
            in_chans, embed_dim, kernel_size=patch_size, stride=patch_size
        )

    def forward(self, x):
        x = self.proj(x).flatten(2).transpose(1, 2)
        return x


def get_sinusoid_encoding(n_position, d_hid):
    """Sinusoid position encoding table"""

    def get_position_angle_vec(position):
        return [
            position / np.power(10000, 2 * (hid_j // 2) / d_hid)
            for hid_j in range(d_hid)
        ]

    sinusoid_table = np.array(
        [get_position_angle_vec(pos_i) for pos_i in range(n_position)]
    )
    sinusoid_table[:, 0::2] = np.sin(sinusoid_table[:, 0::2])  # dim 2i
    sinusoid_table[:, 1::2] = np.cos(sinusoid_table[:, 1::2])  # dim 2i+1

    return torch.FloatTensor(sinusoid_table).unsqueeze(0)


class ASTModel(nn.Module):
    def __init__(
        self,
        label_dim=527,
        fshape=128,
        tshape=2,
        fstride=128,
        tstride=2,
        input_fdim=128,
        input_tdim=1024,
        model_size="base",
        pretrain_stage=True,
        load_pretrained_mdl_path=None,
    ):
        self.num_vertical_patch = (input_fdim - fshape) // fstride + 1

        super(ASTModel, self).__init__()
        assert (
            timm.__version__ == "0.4.5"
        ), "Please use timm == 0.4.5, the code might not be compatible with newer versions."

        # override timm input shape restriction
        timm.models.vision_transformer.PatchEmbed = PatchEmbed

        # pretrain the AST models
        if pretrain_stage == True:
            if load_pretrained_mdl_path != None:
                raise ValueError(
                    "Setting load_pretrained_mdl_path at pretraining stage is useless, pretraining is always from scratch, please change it to None."
                )
            if fstride != fshape or tstride != tshape:
                raise ValueError(
                    "fstride != fshape or tstride != tshape, they must be same at the pretraining stage, patch split overlapping is not supported."
                )

            # if AudioSet pretraining is not used (but ImageNet pretraining may still apply)
            if model_size == "tiny":
                self.v = timm.create_model(
                    "vit_deit_tiny_distilled_patch16_224", pretrained=False
                )
                self.heads, self.depth = 3, 12
                self.cls_token_num = 2
            elif model_size == "small":
                self.v = timm.create_model(
                    "vit_deit_small_distilled_patch16_224", pretrained=False
                )
                self.heads, self.depth = 6, 12
                self.cls_token_num = 2
            elif model_size == "base":
                self.v = timm.create_model(
                    "vit_deit_base_distilled_patch16_384", pretrained=False
                )
                self.heads, self.depth = 12, 12
                self.cls_token_num = 2
            elif model_size == "base_nokd":
                self.v = timm.create_model(
                    "vit_deit_base_patch16_384", pretrained=False
                )
                self.heads, self.depth = 12, 12
                self.cls_token_num = 1
            else:
                raise Exception(
                    "Model size must be one of tiny, small, base, base_nokd"
                )

            self.original_num_patches = self.v.patch_embed.num_patches
            self.oringal_hw = int(self.original_num_patches**0.5)
            self.original_embedding_dim = self.v.pos_embed.shape[2]

            # SSL Pretraining Code
            self.softmax = nn.Softmax(dim=-1)
            self.lsoftmax = nn.LogSoftmax(dim=-1)
            self.fshape, self.tshape = fshape, tshape
            self.fstride, self.tstride = fstride, tstride
            self.input_fdim, self.input_tdim = input_fdim, input_tdim
            # this is a trick to make state_dict to track pretraining input_fdim and input_tdim and save them by using torch.save
            self.p_input_fdim, self.p_input_tdim = nn.Parameter(
                torch.tensor(input_fdim), requires_grad=False
            ), nn.Parameter(torch.tensor(input_tdim), requires_grad=False)

            # masked patch classification (discriminative objective) layer
            # we use two layers for pretext task, but using a single layer has similar performance.
            # we map the output of transformer (768-dim for base models) to 256-dim patch input space, and then dot product with flattened patch input (also 256-dim) to calculate loss.
            # alternatively, you can map the output of transformer to 768-dim patch embedding space, and dot product with patch embedding. Performance-wise they are similar, but map to 256 space is more efficient.
            self.cpredlayer = nn.Sequential(
                nn.Linear(self.original_embedding_dim, self.original_embedding_dim),
                nn.ReLU(),
                nn.Linear(self.original_embedding_dim, 256),
            )
            # masked patch reconstruction (generative objective) layer
            self.gpredlayer = nn.Sequential(
                nn.Linear(self.original_embedding_dim, self.original_embedding_dim),
                nn.ReLU(),
                nn.Linear(self.original_embedding_dim, 256),
            )
            self.unfold = torch.nn.Unfold(
                kernel_size=(fshape, tshape), stride=(fstride, tstride)
            )

            # we use learnable mask embedding (follow the BEIT paper), but using a fixed mask embedding (e.g., 0) leads to same performance.
            self.mask_embed = nn.Parameter(
                torch.zeros([1, 1, self.original_embedding_dim])
            )
            self.mask_embed = torch.nn.init.xavier_normal_(self.mask_embed)

            # get the intermediate shape
            self.p_f_dim, self.p_t_dim = self.get_shape(
                fstride, tstride, input_fdim, input_tdim, fshape, tshape
            )
            num_patches = self.p_f_dim * self.p_t_dim
            self.num_patches = num_patches
            self.v.patch_embed.num_patches = num_patches
            logger.info(
                "pretraining patch split stride: frequency={:d}, time={:d}".format(
                    fstride, tstride
                )
            )
            logger.info(
                "pretraining patch shape: frequency={:d}, time={:d}".format(
                    fshape, tshape
                )
            )
            logger.info(
                "pretraining patch array dimension: frequency={:d}, time={:d}".format(
                    self.p_f_dim, self.p_t_dim
                )
            )
            logger.info("pretraining number of patches={:d}".format(num_patches))

            # the linear patch projection layer, use 1 channel for spectrogram rather than the original 3 channels for RGB images.
            new_proj = torch.nn.Conv2d(
                1,
                self.original_embedding_dim,
                kernel_size=(fshape, tshape),
                stride=(fstride, tstride),
            )
            self.v.patch_embed.proj = new_proj

            # use trainable positional embedding
            new_pos_embed = nn.Parameter(
                torch.zeros(
                    1,
                    self.v.patch_embed.num_patches + self.cls_token_num,
                    self.original_embedding_dim,
                )
            )
            self.v.pos_embed = new_pos_embed
            trunc_normal_(self.v.pos_embed, std=0.02)

        # use a pretrained models for finetuning
        elif pretrain_stage == False:
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            if load_pretrained_mdl_path == None:
                raise ValueError(
                    "Please set load_pretrained_mdl_path to load a pretrained models."
                )
            sd = torch.load(load_pretrained_mdl_path, map_location=device)
            # get the fshape and tshape, input_fdim and input_tdim in the pretraining stage
            try:
                p_fshape, p_tshape = (
                    sd["module.v.patch_embed.proj.weight"].shape[2],
                    sd["module.v.patch_embed.proj.weight"].shape[3],
                )
                p_input_fdim, p_input_tdim = (
                    sd["module.p_input_fdim"].item(),
                    sd["module.p_input_tdim"].item(),
                )
            except:
                raise ValueError(
                    "The model loaded is not from a torch.nn.Dataparallel object. Wrap it with torch.nn.Dataparallel and try again."
                )

            logger.info(
                "now load a SSL pretrained models from " + load_pretrained_mdl_path
            )
            # during pretraining, fstride=fshape and tstride=tshape because no patch overlapping is used
            # here, input_fdim and input_tdim should be that used in pretraining, not that in the fine-tuning.
            # we need to know input_fdim and input_tdim to do positional embedding cut/interpolation.
            # generally it should be better to use same input_fdim during pretraining and finetuning, but input_tdim can be safely different
            audio_model = ASTModel(
                fstride=p_fshape,
                tstride=p_tshape,
                fshape=p_fshape,
                tshape=p_tshape,
                input_fdim=p_input_fdim,
                input_tdim=p_input_tdim,
                pretrain_stage=True,
                model_size=model_size,
            )
            audio_model = torch.nn.DataParallel(audio_model)
            audio_model.load_state_dict(sd, strict=False)

            self.v = audio_model.module.v
            self.original_embedding_dim = self.v.pos_embed.shape[2]
            self.cls_token_num = audio_model.module.cls_token_num

            # mlp head for fine-tuning
            self.mlp_head = nn.Sequential(
                nn.LayerNorm(self.original_embedding_dim),
                nn.Linear(self.original_embedding_dim, label_dim),
            )

            f_dim, t_dim = self.get_shape(
                fstride, tstride, input_fdim, input_tdim, fshape, tshape
            )
            # patch array dimension during pretraining
            p_f_dim, p_t_dim = audio_model.module.p_f_dim, audio_model.module.p_t_dim
            num_patches = f_dim * t_dim
            p_num_patches = p_f_dim * p_t_dim
            self.v.patch_embed.num_patches = num_patches
            logger.info(
                "fine-tuning patch split stride: frequncey={:d}, time={:d}".format(
                    fstride, tstride
                )
            )
            logger.info("fine-tuning number of patches={:d}".format(num_patches))

            # patch shape should be same for pretraining and fine-tuning
            if fshape != p_fshape or tshape != p_tshape:
                raise ValueError(
                    "The patch shape of pretraining and fine-tuning is not consistant, pretraining: f={:d}, t={:d}, finetuning: f={:d}, t={:d}".format(
                        p_fshape, p_tshape, fshape, tshape
                    )
                )

            # patch split stride generally should be different for pretraining and fine-tuning, as patch split overlapping is only used in finetuning
            # during pretraining, p_fshape = p_fstride and p_tshape = p_tstride
            if fstride != p_fshape or tstride != p_tshape:
                # initialize a new patch embedding layer with desired new stride.
                new_proj = torch.nn.Conv2d(
                    1,
                    self.original_embedding_dim,
                    kernel_size=(fshape, tshape),
                    stride=(fstride, tstride),
                )
                # but the weights of patch embedding layer is still got from the pretrained models
                new_proj.weight = torch.nn.Parameter(
                    torch.sum(self.v.patch_embed.proj.weight, dim=1).unsqueeze(1)
                )
                new_proj.bias = self.v.patch_embed.proj.bias
                self.v.patch_embed.proj = new_proj

            new_pos_embed = (
                self.v.pos_embed[:, self.cls_token_num :, :]
                .detach()
                .reshape(1, p_num_patches, self.original_embedding_dim)
                .transpose(1, 2)
                .reshape(1, self.original_embedding_dim, p_f_dim, p_t_dim)
            )
            # cut or interpolate the positional embedding
            if t_dim < p_t_dim:
                new_pos_embed = new_pos_embed[
                    :,
                    :,
                    :,
                    int(p_t_dim / 2)
                    - int(t_dim / 2) : int(p_t_dim / 2)
                    - int(t_dim / 2)
                    + t_dim,
                ]
            else:
                new_pos_embed = torch.nn.functional.interpolate(
                    new_pos_embed, size=(8, t_dim), mode="bilinear"
                )
            if f_dim < p_f_dim:
                new_pos_embed = new_pos_embed[
                    :,
                    :,
                    int(p_f_dim / 2)
                    - int(f_dim / 2) : int(p_f_dim / 2)
                    - int(f_dim / 2)
                    + t_dim,
                    :,
                ]
            else:
                new_pos_embed = torch.nn.functional.interpolate(
                    new_pos_embed, size=(f_dim, t_dim), mode="bilinear"
                )

            new_pos_embed = new_pos_embed.reshape(
                1, self.original_embedding_dim, num_patches
            ).transpose(1, 2)
            self.v.pos_embed = nn.Parameter(
                torch.cat(
                    [
                        self.v.pos_embed[:, : self.cls_token_num, :].detach(),
                        new_pos_embed,
                    ],
                    dim=1,
                )
            )

    # get the shape of intermediate representation.
    def get_shape(self, fstride, tstride, input_fdim, input_tdim, fshape, tshape):
        test_input = torch.randn(1, 1, input_fdim, input_tdim)
        test_proj = nn.Conv2d(
            1,
            self.original_embedding_dim,
            kernel_size=(fshape, tshape),
            stride=(fstride, tstride),
        )
        test_out = test_proj(test_input)
        f_dim = test_out.shape[2]
        t_dim = test_out.shape[3]
        return f_dim, t_dim

    def forward(self, x):
        # expect input x = (batch_size, time_frame_num, frequency_bins), e.g., (12, 1024, 128)
        x = x.unsqueeze(1)
        x = x.transpose(2, 3)

        B = x.shape[0]
        x = self.v.patch_embed(x)
        cls_tokens = self.v.cls_token.expand(B, -1, -1)
        dist_token = self.v.dist_token.expand(B, -1, -1)
        x = torch.cat((cls_tokens, dist_token, x), dim=1)
        x = x + self.v.pos_embed
        x = self.v.pos_drop(x)

        # superb needs represents of each layer
        reps = []
        for blk_id, blk in enumerate(self.v.blocks):
            x = blk(x)

            post_x = x[:, 2:, :]
            post_x = (
                post_x.reshape(B, self.num_vertical_patch, -1, x.size(-1))
                .transpose(1, 2)
                .flatten(2)
            )
            # (batch_size, num_horizon_patch, num_vertical_patch * hidden_size)

            reps.append(post_x)

        x = self.v.norm(x)

        # reps=representation of each layer, x=final representation
        return reps, x



import os

from s3prl.util.download import _urls_to_filepaths

from .expert import UpstreamExpert as _UpstreamExpert


def lighthubert_local(ckpt, *args, **kwargs):
    """
    The model from local ckpt
        ckpt (str): PATH
    """
    assert os.path.isfile(ckpt)
    return _UpstreamExpert(ckpt, *args, **kwargs)


def lighthubert_url(ckpt, refresh=False, *args, **kwargs):
    """
    The model from google drive id
        ckpt (str): URL
        refresh (bool): whether to download ckpt/config again if existed
    """
    return lighthubert_local(_urls_to_filepaths(ckpt, refresh=refresh), *args, **kwargs)


def lighthubert(refresh=False, *args, **kargs):
    """
    The default model - Small
    refresh (bool): whether to download ckpt/config again if existed
    """
    return lighthubert_small(refresh=refresh, *args, **kargs)


def lighthubert_small(refresh=False, *args, **kwargs):
    """
    The small model
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs[
        "ckpt"
    ] = "https://huggingface.co/mechanicalsea/lighthubert/resolve/main/lighthubert_small.pt"
    return lighthubert_url(refresh=refresh, *args, **kwargs)


def lighthubert_base(refresh=False, *args, **kwargs):
    """
    The Base model
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs[
        "ckpt"
    ] = "https://huggingface.co/mechanicalsea/lighthubert/resolve/main/lighthubert_base.pt"
    return lighthubert_url(refresh=refresh, *args, **kwargs)


def lighthubert_stage1(refresh=False, *args, **kwargs):
    """
    The Stage1 model
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs[
        "ckpt"
    ] = "https://huggingface.co/mechanicalsea/lighthubert/resolve/main/lighthubert_stage1.pt"
    return lighthubert_url(refresh=refresh, *args, **kwargs)



import json
from collections import defaultdict
from copy import deepcopy
from pathlib import Path

import pandas as pd
import torchaudio
from omegaconf import MISSING

from ._hear_util import resample_hear_corpus
from .hear_dcase_2016_task2 import HearDcase2016Task2

MAESTRO_NUM_FOLDS = 5

__all__ = ["HearMaestro"]


def prepare_maestro(
    target_dir: str,
    cache_dir: str,
    dataset_root: str,
    test_fold: int = 0,
    get_path_only: bool = False,
):
    target_dir: Path = Path(target_dir)
    train_csv = target_dir / "train.csv"
    valid_csv = target_dir / "valid.csv"
    test_csv = target_dir / "test.csv"
    if get_path_only:
        return train_csv, valid_csv, [test_csv]

    assert test_fold < MAESTRO_NUM_FOLDS, (
        f"MAESTRO only has {MAESTRO_NUM_FOLDS} folds but get 'test_fold' "
        f"arguments {test_fold}"
    )

    resample_hear_corpus(dataset_root, target_sr=16000)

    dataset_root = Path(dataset_root)
    wav_root = dataset_root / "16000"

    NUM_FOLD = 5
    test_id = test_fold
    valid_id = (test_fold + 1) % NUM_FOLD
    train_ids = [idx for idx in range(NUM_FOLD) if idx not in [test_id, valid_id]]

    fold_metas = []
    fold_dfs = []
    for fold_id in range(NUM_FOLD):
        with open(dataset_root / f"fold{fold_id:2d}.json".replace(" ", "0")) as f:
            metadata = json.load(f)
            fold_metas.append(metadata)

        data = defaultdict(list)
        for utt in metadata:
            wav_path = (
                wav_root / f"fold{fold_id:2d}".replace(" ", "0") / utt
            ).resolve()
            info = torchaudio.info(wav_path)
            baseinfo = {
                "record_id": utt,
                "wav_path": str(wav_path),
                "duration": info.num_frames / info.sample_rate,
            }
            for segment in metadata[utt]:
                fullinfo = deepcopy(baseinfo)
                fullinfo[
                    "utt_id"
                ] = f"{baseinfo['record_id']}-{int(segment['start'])}-{int(segment['end'])}"
                fullinfo["labels"] = segment["label"]
                fullinfo["start_sec"] = segment["start"] / 1000
                fullinfo["end_sec"] = segment["end"] / 1000

                for key, value in fullinfo.items():
                    data[key].append(value)

        fold_dfs.append(pd.DataFrame(data=data))

    test_meta, test_data = fold_metas[test_id], fold_dfs[test_id]
    valid_meta, valid_data = fold_metas[valid_id], fold_dfs[valid_id]
    train_meta, train_data = {}, []
    for idx in train_ids:
        train_meta.update(fold_metas[idx])
        train_data.append(fold_dfs[idx])
    train_data: pd.DataFrame = pd.concat(train_data)

    train_data.to_csv(train_csv, index=False)
    valid_data.to_csv(valid_csv, index=False)
    test_data.to_csv(test_csv, index=False)

    return train_csv, valid_csv, [test_csv]


class HearMaestro(HearDcase2016Task2):
    def default_config(self) -> dict:
        return dict(
            start=0,
            stop=None,
            target_dir=MISSING,
            cache_dir=None,
            remove_all_cache=False,
            prepare_data=dict(
                dataset_root=MISSING,
                test_fold=MISSING,
            ),
            build_batch_sampler=dict(
                train=dict(
                    batch_size=5,
                    shuffle=True,
                ),
                valid=dict(
                    item="record_id",
                ),
                test=dict(
                    item="record_id",
                ),
            ),
            build_upstream=dict(
                name=MISSING,
            ),
            build_featurizer=dict(
                layer_selections=None,
                normalize=False,
            ),
            build_downstream=dict(
                hidden_layers=2,
            ),
            build_model=dict(
                upstream_trainable=False,
            ),
            build_task=dict(
                prediction_type="multilabel",
                scores=["event_onset_50ms_fms", "event_onset_offset_50ms_20perc_fms"],
                postprocessing_grid={
                    "median_filter_ms": [150],
                    "min_duration": [50],
                },
            ),
            build_optimizer=dict(
                name="Adam",
                conf=dict(
                    lr=1.0e-3,
                ),
            ),
            build_scheduler=dict(
                name="ExponentialLR",
                gamma=0.9,
            ),
            save_model=dict(),
            save_task=dict(),
            train=dict(
                total_steps=15000,
                log_step=100,
                eval_step=500,
                save_step=500,
                gradient_clipping=1.0,
                gradient_accumulate=1,
                valid_metric="event_onset_50ms_fms",
                valid_higher_better=True,
                auto_resume=True,
                resume_ckpt_dir=None,
            ),
            evaluate=dict(),
        )

    def prepare_data(
        self,
        prepare_data: dict,
        target_dir: str,
        cache_dir: str,
        get_path_only: bool = False,
    ):
        return prepare_maestro(
            **self._get_current_arguments(flatten_dict="prepare_data")
        )



"""
The abstract Task

Authors
  * Leo 2022
"""

import abc
from collections import defaultdict
from typing import List

import torch

__all__ = ["Task"]


class Task(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def get_state(self):
        # self.model will be separately saved, do not save self.model.state_dict() here
        return {}

    def set_state(self, state: dict):
        pass

    def parse_cached_results(self, cached_results: List[dict]):
        keys = list(cached_results[0].keys())
        dol = defaultdict(list)
        for d in cached_results:
            assert sorted(keys) == sorted(list(d.keys()))
            for k, v in d.items():
                if isinstance(v, (tuple, list)):
                    dol[k].extend(v)
                else:
                    dol[k].append(v)
        return dict(dol)

    @abc.abstractmethod
    def predict(self):
        raise NotImplementedError

    def forward(self, mode: str, *args, **kwargs):
        return getattr(self, f"{mode}_step")(*args, **kwargs)

    def reduction(self, mode: str, *args, **kwargs):
        return getattr(self, f"{mode}_reduction")(*args, **kwargs)

    @abc.abstractmethod
    def train_step(self):
        raise NotImplementedError

    @abc.abstractmethod
    def valid_step(self):
        raise NotImplementedError

    @abc.abstractmethod
    def test_step(self):
        raise NotImplementedError

    @abc.abstractmethod
    def train_reduction(self):
        raise NotImplementedError

    @abc.abstractmethod
    def valid_reduction(self):
        raise NotImplementedError

    @abc.abstractmethod
    def test_reduction(self):
        raise NotImplementedError



"""
The most commonly used batch sampler, recover the default batch sampler used
in torch DataLoader

Authors:
  * Leo 2022
"""

import torch
from torch.utils.data import BatchSampler, RandomSampler, SequentialSampler

__all__ = ["FixedBatchSizeBatchSampler"]


class FixedBatchSizeBatchSampler:
    """
    The reduced timestamps for a batch should not exceed the max_timestamp.
    If shuffled, each indices are first shuffled before aggregated into batches

    Args:
        data_source: __len__ is implemented
    """

    def __init__(
        self,
        data_source,
        batch_size: int,
        shuffle: bool = False,
        seed: int = 12345678,
    ) -> None:
        self.batch_size = batch_size
        self.seed = seed
        self.shuffle = shuffle

        if shuffle:
            self.generator = torch.Generator()
            self.sampler = RandomSampler(data_source, generator=self.generator)
        else:
            self.sampler = SequentialSampler(data_source)

    def set_epoch(self, epoch: int) -> None:
        if self.shuffle:
            self.generator.manual_seed(self.seed + epoch)

    def _evaluate_reduced_timestamps(self, batch_indices):
        return self.reduce_func([self.timestamps[indice] for indice in batch_indices])

    def __iter__(self):
        batch_sampler = BatchSampler(
            self.sampler, batch_size=self.batch_size, drop_last=False
        )
        return iter(batch_sampler)

    def __len__(self):
        return len(list(iter(self)))



from .base import SequentialDataPipe
from .common_pipes import (
    EncodeCategory,
    EncodeMultiLabel,
    EncodeMultipleCategory,
    LoadAudio,
    SetOutputKeys,
)


class UtteranceClassificationPipe(SequentialDataPipe):
    """
    each item in the input dataset should have:
        wav_path: str
        label: str
    """

    def __init__(
        self,
        output_keys: dict = None,
        audio_sample_rate: int = 16000,
        audio_channel_reduction: str = "first",
        sox_effects: list = None,
        train_category_encoder: bool = False,
    ):
        output_keys = output_keys or dict(
            x="wav",
            x_len="wav_len",
            class_id="class_id",
            label="label",
            unique_name="id",
        )

        super().__init__(
            LoadAudio(
                audio_sample_rate=audio_sample_rate,
                audio_channel_reduction=audio_channel_reduction,
                sox_effects=sox_effects,
            ),
            EncodeCategory(train_category_encoder=train_category_encoder),
            SetOutputKeys(output_keys=output_keys),
        )


class UtteranceMultipleCategoryClassificationPipe(SequentialDataPipe):
    """
    each item in the input dataset should have:
        wav_path: str
        labels: List[str]
    """

    def __init__(
        self,
        output_keys: dict = None,
        audio_sample_rate: int = 16000,
        audio_channel_reduction: str = "first",
        sox_effects: list = None,
        train_category_encoder: bool = False,
    ):
        output_keys = output_keys or dict(
            x="wav",
            x_len="wav_len",
            class_ids="class_ids",
            labels="labels",
            unique_name="id",
        )

        super().__init__(
            LoadAudio(
                audio_sample_rate=audio_sample_rate,
                audio_channel_reduction=audio_channel_reduction,
                sox_effects=sox_effects,
            ),
            EncodeMultipleCategory(train_category_encoder=train_category_encoder),
            SetOutputKeys(output_keys=output_keys),
        )


class HearScenePipe(SequentialDataPipe):
    """
    each item in the input dataset should have:
        wav_path: str
        labels: List[str]
    """

    def __init__(
        self,
        output_keys: dict = None,
        audio_sample_rate: int = 16000,
        audio_channel_reduction: str = "first",
    ):
        output_keys = output_keys or dict(
            x="wav",
            x_len="wav_len",
            y="binary_labels",
            labels="labels",
            unique_name="id",
        )

        super().__init__(
            LoadAudio(
                audio_sample_rate=audio_sample_rate,
                audio_channel_reduction=audio_channel_reduction,
            ),
            EncodeMultiLabel(),
            SetOutputKeys(output_keys=output_keys),
        )



"""BYOL for Audio: Dataset class definition."""

import random

import librosa
import numpy as np
import torch
import torch.nn.functional as F
import torchaudio
from torch.utils.data import Dataset


class MelSpectrogramLibrosa:
    """Mel spectrogram using librosa."""

    def __init__(self, fs=16000, n_fft=1024, shift=160, n_mels=64, fmin=60, fmax=7800):
        self.fs, self.n_fft, self.shift, self.n_mels, self.fmin, self.fmax = (
            fs,
            n_fft,
            shift,
            n_mels,
            fmin,
            fmax,
        )
        self.mfb = librosa.filters.mel(
            sr=fs, n_fft=n_fft, n_mels=n_mels, fmin=fmin, fmax=fmax
        )

    def __call__(self, audio):
        X = librosa.stft(np.array(audio), n_fft=self.n_fft, hop_length=self.shift)
        return torch.tensor(np.matmul(self.mfb, np.abs(X) ** 2 + np.finfo(float).eps))


class WaveInLMSOutDataset(Dataset):
    """Wave in, log-mel spectrogram out, dataset class.

    Choosing librosa or torchaudio:
        librosa: Stable but slower.
        torchaudio: Faster but cannot reproduce the exact performance of pretrained weight,
            which might be caused by the difference with librosa. Librosa was used in the pretraining.

    Args:
        cfg: Configuration settings.
        audio_files: List of audio file pathnames.
        labels: List of labels corresponding to the audio files.
        transform: Transforms (augmentations), callable.
        use_librosa: True if using librosa for converting audio to log-mel spectrogram (LMS).
    """

    def __init__(self, cfg, audio_files, labels, transform, use_librosa=False):
        # argment check
        assert (labels is None) or (
            len(audio_files) == len(labels)
        ), "The number of audio files and labels has to be the same."
        super().__init__()

        # initializations
        self.cfg = cfg
        self.files = audio_files
        self.labels = labels
        self.transform = transform
        self.unit_length = int(cfg.unit_sec * cfg.sample_rate)
        self.to_melspecgram = (
            MelSpectrogramLibrosa(
                fs=cfg.sample_rate,
                n_fft=cfg.n_fft,
                shift=cfg.hop_length,
                n_mels=cfg.n_mels,
                fmin=cfg.f_min,
                fmax=cfg.f_max,
            )
            if use_librosa
            else torchaudio.transforms.MelSpectrogram(
                sample_rate=cfg.sample_rate,
                n_fft=cfg.n_fft,
                win_length=cfg.win_length,
                hop_length=cfg.hop_length,
                n_mels=cfg.n_mels,
                f_min=cfg.f_min,
                f_max=cfg.f_max,
                power=2,
            )
        )

    def __len__(self):
        return len(self.files)

    def __getitem__(self, idx):
        # load single channel .wav audio
        # print(self.files[idx])
        try:
            wav, sr = torchaudio.load(self.files[idx])
        except RuntimeError:
            print(self.files[idx])
            raise FileNotFoundError(self.files[idx])
        assert (
            sr == self.cfg.sample_rate
        ), f"Convert .wav files to {self.cfg.sample_rate} Hz. {self.files[idx]} has {sr} Hz."
        assert (
            wav.shape[0] == 1
        ), f"Convert .wav files to single channel audio, {self.files[idx]} has {wav.shape[0]} channels."
        wav = wav[0]  # (1, length) -> (length,)

        # zero padding to both ends
        length_adj = self.unit_length - len(wav)
        if length_adj > 0:
            half_adj = length_adj // 2
            wav = F.pad(wav, (half_adj, length_adj - half_adj))

        # random crop unit length wave
        length_adj = self.unit_length - len(wav)
        start = random.randint(0, length_adj) if length_adj > 0 else 0
        wav = wav[start : start + self.unit_length]

        # to log mel spectrogram -> (1, n_mels, time)
        lms = (self.to_melspecgram(wav) + torch.finfo().eps).log().unsqueeze(0)

        # transform (augment)
        if self.transform:
            lms = self.transform(lms)

        if self.labels is not None:
            return lms, torch.tensor(self.labels[idx])
        return lms



import torch

from s3prl.dataio.dataset.load_audio import LoadAudio
from s3prl.util.pseudo_data import pseudo_audio


def test_load_audio():
    with pseudo_audio([3.0, 4.0, 5.2]) as (paths, num_frames):
        dataset = LoadAudio(paths, [None, 1.0, 3.1], [None, 3.2, None], max_secs=4.2)

        for item in dataset:
            assert isinstance(item["wav"], torch.Tensor)



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/byol_a/hubconf.py ]
#   Synopsis     [ the BYOL-A torch hubconf ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""

from pathlib import Path as _Path

from s3prl.util.download import _urls_to_filepaths

from .expert import UpstreamExpert as _UpstreamExpert

DEFAULT_CONFIG_PATH = _Path(__file__).parent / "config.yaml"


def byol_a_2048(refresh=False, **kwds):
    ckpt = _urls_to_filepaths(
        "https://github.com/nttcslab/byol-a/raw/master/pretrained_weights/AudioNTT2020-BYOLA-64x96d2048.pth",
        refresh=refresh,
    )
    return _UpstreamExpert(ckpt, DEFAULT_CONFIG_PATH, 2048, **kwds)


def byol_a_1024(refresh=False, **kwds):
    ckpt = _urls_to_filepaths(
        "https://github.com/nttcslab/byol-a/raw/master/pretrained_weights/AudioNTT2020-BYOLA-64x96d1024.pth",
        refresh=refresh,
    )
    return _UpstreamExpert(ckpt, DEFAULT_CONFIG_PATH, 1024, **kwds)


def byol_a_512(refresh=False, **kwds):
    ckpt = _urls_to_filepaths(
        "https://github.com/nttcslab/byol-a/raw/master/pretrained_weights/AudioNTT2020-BYOLA-64x96d512.pth",
        refresh=refresh,
    )
    return _UpstreamExpert(ckpt, DEFAULT_CONFIG_PATH, 512, **kwds)



import logging

import pytest

from s3prl.dataio.sampler import (
    DistributedBatchSamplerWrapper,
    FixedBatchSizeBatchSampler,
    MaxTimestampBatchSampler,
)

logger = logging.getLogger(__name__)


def _merge_batch_indices(batch_indices):
    all_indices = []
    for indices in batch_indices:
        all_indices += indices
    return all_indices


@pytest.mark.parametrize("world_size", [1, 2, 3, 4, 5, 6, 7, 8])
def test_distributed_sampler(world_size):
    sampler = [[1, 2, 3], [4, 5, 6, 7], [8], [9, 10]]
    ddp_indices = []
    for rank in range(world_size):
        ddp_sampler = DistributedBatchSamplerWrapper(sampler, world_size, rank)
        ddp_indices += _merge_batch_indices(ddp_sampler)
    assert sorted(ddp_indices) == sorted(_merge_batch_indices(sampler))


timestamps = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
data = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]


@pytest.mark.parametrize("batch_size", [1, 2, 3, len(data)])
def test_FixedBatchSizeBatchSampler(batch_size):
    dataset = data
    iter1 = list(iter(FixedBatchSizeBatchSampler(dataset, batch_size, shuffle=False)))
    iter2 = list(iter(FixedBatchSizeBatchSampler(dataset, batch_size, shuffle=True)))
    indices1 = sorted(_merge_batch_indices(iter1))
    indices2 = sorted(_merge_batch_indices(iter2))
    assert indices1 == indices2 == list(range(len(timestamps)))



import copy

import torch
import torch.nn as nn

from s3prl import Output
from s3prl.nn.vq_apc import VqApcLayer


class MaskConvBlock(nn.Module):
    """
    Masked Convolution Blocks as described in NPC paper
    """

    def __init__(self, input_size, hidden_size, kernel_size, mask_size):
        super(MaskConvBlock, self).__init__()
        assert kernel_size - mask_size > 0, "Mask > kernel somewhere in the model"
        # CNN for computing feature (ToDo: other activation?)
        self.act = nn.Tanh()
        self.pad_size = (kernel_size - 1) // 2
        self.conv = nn.Conv1d(
            in_channels=input_size,
            out_channels=hidden_size,
            kernel_size=kernel_size,
            padding=self.pad_size,
        )
        # Fixed mask for NPC
        mask_head = (kernel_size - mask_size) // 2
        mask_tail = mask_head + mask_size
        conv_mask = torch.ones_like(self.conv.weight)
        conv_mask[:, :, mask_head:mask_tail] = 0
        self.register_buffer("conv_mask", conv_mask)

    def forward(self, feat):
        feat = nn.functional.conv1d(
            feat,
            self.conv_mask * self.conv.weight,
            bias=self.conv.bias,
            padding=self.pad_size,
        )
        feat = feat.permute(0, 2, 1)  # BxCxT -> BxTxC
        feat = self.act(feat)
        return feat


class ConvBlock(nn.Module):
    """
    Convolution Blocks as described in NPC paper
    """

    def __init__(
        self, input_size, hidden_size, residual, dropout, batch_norm, activate
    ):
        super(ConvBlock, self).__init__()
        self.residual = residual
        if activate == "relu":
            self.act = nn.ReLU()
        elif activate == "tanh":
            self.act = nn.Tanh()
        else:
            raise NotImplementedError
        self.conv = nn.Conv1d(
            input_size, hidden_size, kernel_size=3, stride=1, padding=1
        )
        self.linear = nn.Conv1d(
            hidden_size, hidden_size, kernel_size=1, stride=1, padding=0
        )
        self.batch_norm = batch_norm
        if batch_norm:
            self.bn1 = nn.BatchNorm1d(hidden_size)
            self.bn2 = nn.BatchNorm1d(hidden_size)
        self.dropout = nn.Dropout(dropout)

    def forward(self, feat):
        res = feat
        out = self.conv(feat)
        if self.batch_norm:
            out = self.bn1(out)
        out = self.act(out)
        out = self.linear(out)
        if self.batch_norm:
            out = self.bn2(out)
        out = self.dropout(out)
        if self.residual:
            out = out + res
        return self.act(out)


class CnnNpc(nn.Module):
    """
    The NPC model with stacked ConvBlocks & Masked ConvBlocks
    """

    def __init__(
        self,
        input_size,
        hidden_size,
        n_blocks,
        dropout,
        residual,
        kernel_size,
        mask_size,
        vq=None,
        batch_norm=True,
        activate="relu",
        disable_cross_layer=False,
        dim_bottleneck=None,
    ):
        super(CnnNpc, self).__init__()

        # Setup
        assert kernel_size % 2 == 1, "Kernel size can only be odd numbers"
        assert mask_size % 2 == 1, "Mask size can only be odd numbers"
        assert n_blocks >= 1, "At least 1 block needed"
        self.code_dim = hidden_size
        self.n_blocks = n_blocks
        self.input_mask_size = mask_size
        self.kernel_size = kernel_size
        self.disable_cross_layer = disable_cross_layer
        self.apply_vq = vq is not None
        self.apply_ae = dim_bottleneck is not None
        if self.apply_ae:
            assert not self.apply_vq
            self.dim_bottleneck = dim_bottleneck

        # Build blocks
        self.blocks, self.masked_convs = [], []
        cur_mask_size = mask_size
        for i in range(n_blocks):
            h_dim = input_size if i == 0 else hidden_size
            res = False if i == 0 else residual
            # ConvBlock
            self.blocks.append(
                ConvBlock(h_dim, hidden_size, res, dropout, batch_norm, activate)
            )
            # Masked ConvBlock on each or last layer
            cur_mask_size = cur_mask_size + 2
            if self.disable_cross_layer and (i != (n_blocks - 1)):
                self.masked_convs.append(None)
            else:
                self.masked_convs.append(
                    MaskConvBlock(hidden_size, hidden_size, kernel_size, cur_mask_size)
                )
        self.blocks = nn.ModuleList(self.blocks)
        self.masked_convs = nn.ModuleList(self.masked_convs)

        # Creates N-group VQ
        if self.apply_vq:
            self.vq_layers = []
            vq_config = copy.deepcopy(vq)
            codebook_size = vq_config.pop("codebook_size")
            self.vq_code_dims = vq_config.pop("code_dim")
            assert len(self.vq_code_dims) == len(codebook_size)
            assert sum(self.vq_code_dims) == hidden_size
            for cs, cd in zip(codebook_size, self.vq_code_dims):
                self.vq_layers.append(
                    VqApcLayer(
                        input_size=cd, code_dim=cd, codebook_size=cs, **vq_config
                    )
                )
            self.vq_layers = nn.ModuleList(self.vq_layers)

        # Back to spectrogram
        if self.apply_ae:
            self.ae_bottleneck = nn.Linear(hidden_size, self.dim_bottleneck, bias=False)
            self.postnet = nn.Linear(self.dim_bottleneck, input_size)
        else:
            self.postnet = nn.Linear(hidden_size, input_size)

    def create_msg(self):
        msg_list = []
        msg_list.append(
            "Model spec.| Method = NPC\t| # of Blocks = {}\t".format(self.n_blocks)
        )
        msg_list.append(
            "           | Desired input mask size = {}".format(self.input_mask_size)
        )
        msg_list.append(
            "           | Receptive field size = {}".format(
                self.kernel_size + 2 * self.n_blocks
            )
        )
        return msg_list

    def report_ppx(self):
        """
        Returns perplexity of VQ distribution
        """
        if self.apply_vq:
            # ToDo: support more than 2 groups
            rt = [vq_layer.report_ppx() for vq_layer in self.vq_layers] + [None]
            return rt[0], rt[1]
        else:
            return None, None

    def report_usg(self):
        """
        Returns usage of VQ codebook
        """
        if self.apply_vq:
            # ToDo: support more than 2 groups
            rt = [vq_layer.report_usg() for vq_layer in self.vq_layers] + [None]
            return rt[0], rt[1]
        else:
            return None, None

    def get_unmasked_feat(self, sp_seq, n_layer):
        """
        Returns unmasked features from n-th layer ConvBlock
        """
        unmasked_feat = sp_seq.permute(0, 2, 1)  # BxTxC -> BxCxT
        for i in range(self.n_blocks):
            unmasked_feat = self.blocks[i](unmasked_feat)
            if i == n_layer:
                unmasked_feat = unmasked_feat.permute(0, 2, 1)
                break
        return unmasked_feat

    def forward(self, sp_seq, testing=False):
        # BxTxC -> BxCxT (reversed in Masked ConvBlock)
        unmasked_feat = sp_seq.permute(0, 2, 1)
        # Forward through each layer
        for i in range(self.n_blocks):
            unmasked_feat = self.blocks[i](unmasked_feat)
            if self.disable_cross_layer:
                # Last layer masked feature only
                if i == (self.n_blocks - 1):
                    feat = self.masked_convs[i](unmasked_feat)
            else:
                # Masked feature aggregation
                masked_feat = self.masked_convs[i](unmasked_feat)
                if i == 0:
                    feat = masked_feat
                else:
                    feat = feat + masked_feat
        # Apply bottleneck and predict spectrogram
        if self.apply_vq:
            q_feat = []
            offet = 0
            for vq_layer, cd in zip(self.vq_layers, self.vq_code_dims):
                q_f = vq_layer(feat[:, :, offet : offet + cd], testing).output
                q_feat.append(q_f)
                offet += cd
            q_feat = torch.cat(q_feat, dim=-1)
            pred = self.postnet(q_feat)
        elif self.apply_ae:
            feat = self.ae_bottleneck(feat)
            pred = self.postnet(feat)
        else:
            pred = self.postnet(feat)
        return Output(hidden_states=feat, prediction=pred)



"""
Permutation Invariant Training (PIT) loss

Authors:
  * Jiatong Shi 2021
"""

from itertools import permutations

import numpy as np
import torch

__all__ = [
    "pit_loss",
]


# compute mask to remove the padding positions
def create_length_mask(length, max_len, num_output, device):
    batch_size = len(length)
    mask = torch.zeros(batch_size, max_len, num_output)
    for i in range(batch_size):
        mask[i, : length[i], :] = 1
    mask = mask.to(device)
    return mask


# compute loss for a single permutation
def pit_loss_single_permute(output, label, length):
    bce_loss = torch.nn.BCEWithLogitsLoss(reduction="none")
    mask = create_length_mask(length, label.size(1), label.size(2), label.device)
    loss = bce_loss(output, label)
    loss = loss * mask
    loss = torch.sum(torch.mean(loss, dim=2), dim=1)
    loss = torch.unsqueeze(loss, dim=1)
    return loss


def pit_loss(output, label, length):
    """
    The Permutation Invariant Training loss

    Args:
        output (torch.FloatTensor): prediction in (batch_size, seq_len, num_class)
        label (torch.FloatTensor): label in the same shape as :code:`output`
        length (torch.LongTensor): the valid length of each instance. :code:`output` and :code:`label`
            share the same valid length

    Returns:
        tuple:

        1. loss (torch.FloatTensor)
        2. min_idx (int): the id with the minimum loss
        3. all the permutation
    """
    num_output = label.size(2)
    device = label.device
    permute_list = [np.array(p) for p in permutations(range(num_output))]
    loss_list = []
    for p in permute_list:
        label_perm = label[:, :, p]
        loss_perm = pit_loss_single_permute(output, label_perm, length)
        loss_list.append(loss_perm)
    loss = torch.cat(loss_list, dim=1)
    min_loss, min_idx = torch.min(loss, dim=1)
    loss = torch.sum(min_loss) / torch.sum(length.float().to(device))
    return loss, min_idx, permute_list


def get_label_perm(label, perm_idx, perm_list):
    batch_size = len(perm_idx)
    label_list = []
    for i in range(batch_size):
        label_list.append(label[i, :, perm_list[perm_idx[i]]].data.cpu().numpy())
    return torch.from_numpy(np.array(label_list)).float()



import logging
import os
from pathlib import Path

import torchaudio
from joblib import Parallel, delayed
from librosa.util import find_files
from tqdm import tqdm

logger = logging.getLogger(__name__)


def resample_hear_corpus(task_dir: str, target_sr: int = 16000, num_workers: int = 6):
    """
    Resample audio files in

    ${task_dir}/48000/

    to

    ${task_dir}/${target_sr}/
    """

    task_dir: Path = Path(task_dir)
    target_audio_dir: Path = task_dir / f"{target_sr}"
    if target_audio_dir.is_dir():
        logger.info(f"{target_audio_dir} already exist. Do not need to resample")
        return

    default_audio_dir = task_dir / "48000"
    assert default_audio_dir.exists(), f"{default_audio_dir} not found"

    split_names = os.listdir(default_audio_dir)
    for split_name in sorted(split_names):
        split_dir = default_audio_dir / split_name
        wav_paths = find_files(split_dir)
        tgt_dir = target_audio_dir / split_name
        tgt_dir.mkdir(exist_ok=True, parents=True)

        def resample(wav_path: str):
            wav, sr = torchaudio.load(wav_path)
            if sr != target_sr:
                resampler = torchaudio.transforms.Resample(sr, target_sr)
                wav = resampler(wav)
            torchaudio.save(
                str(tgt_dir / Path(wav_path).name), wav, sample_rate=target_sr
            )

        logger.info(f"Resampling {split_dir} to {tgt_dir}:")
        Parallel(n_jobs=num_workers)(
            delayed(resample)(path) for path in tqdm(wav_paths)
        )



# Copyright (c) Facebook, Inc. All Rights Reserved

# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/wav2vec2/hubconf.py ]
#   Synopsis     [ the wav2vec 2.0 torch hubconf ]
#   Author       [ S3PRL / Kushal Lakhotia ]
"""*********************************************************************************************"""

import logging
import os
import time
from pathlib import Path

from filelock import FileLock

from s3prl.util.download import _urls_to_filepaths

from .convert import load_and_convert_fairseq_ckpt
from .expert import LegacyUpstreamExpert as _LegacyUpstreamExpert
from .expert import UpstreamExpert as _UpstreamExpert

logger = logging.getLogger(__name__)

NEW_ENOUGH_SECS = 2.0


def wav2vec2_custom(
    ckpt: str,
    legacy: bool = False,
    fairseq: bool = False,
    refresh: bool = False,
    **kwargs,
):
    assert not (legacy and fairseq), (
        "The option 'legacy' will directly load a fairseq checkpoint, "
        "while the option 'fairseq' will first convert the fairseq checkpoint to "
        "be fairseq indenpendent and then load the checkpoint. "
        "These two options cannot be used jointly."
    )

    if ckpt.startswith("http"):
        ckpt = _urls_to_filepaths(ckpt, refresh=refresh)

    if fairseq:
        ckpt: Path = Path(ckpt)
        converted_ckpt = ckpt.parent / f"{ckpt.stem}.converted.pt"
        lock_file = Path(str(converted_ckpt) + ".lock")

        logger.info(f"Converting a fairseq checkpoint: {ckpt}")
        logger.info(f"To: {converted_ckpt}")

        with FileLock(str(lock_file)):
            if not converted_ckpt.is_file() or (
                refresh and (time.time() - os.path.getmtime(ckpt)) > NEW_ENOUGH_SECS
            ):
                load_and_convert_fairseq_ckpt(ckpt, converted_ckpt)

        ckpt = converted_ckpt

    assert os.path.isfile(ckpt)
    if legacy:
        return _LegacyUpstreamExpert(ckpt, **kwargs)
    else:
        return _UpstreamExpert(ckpt, **kwargs)


def wav2vec2_local(*args, **kwargs):
    return wav2vec2_custom(*args, **kwargs)


def wav2vec2_url(*args, **kwargs):
    return wav2vec2_custom(*args, **kwargs)


def wav2vec2(refresh=False, *args, **kwargs):
    """
    The default model - Base
        refresh (bool): whether to download ckpt/config again if existed
    """
    return wav2vec2_base_960(refresh=refresh, *args, **kwargs)


def wav2vec2_base_960(refresh=False, legacy=False, **kwargs):
    """
    The Base model
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs["ckpt"] = "https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec_small.pt"
    if not legacy:
        kwargs[
            "ckpt"
        ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/wav2vec_small.pt"
    return wav2vec2_custom(refresh=refresh, legacy=legacy, **kwargs)


def wav2vec2_large_960(refresh=False, legacy=False, **kwargs):
    """
    The Large model trained on LibriSpeech 960 hours of data
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs["ckpt"] = "https://dl.fbaipublicfiles.com/fairseq/wav2vec/libri960_big.pt"
    if not legacy:
        kwargs[
            "ckpt"
        ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/libri960_big.pt"
    return wav2vec2_custom(refresh=refresh, legacy=legacy, **kwargs)


def wav2vec2_large_ll60k(refresh=False, legacy=False, **kwargs):
    """
    The Large model trained on Libri-light 60k hours of data
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs["ckpt"] = "https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec_vox_new.pt"
    if not legacy:
        kwargs[
            "ckpt"
        ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/wav2vec_vox_new.pt"
    return wav2vec2_custom(refresh=refresh, legacy=legacy, **kwargs)


def wav2vec2_large_lv60_cv_swbd_fsh(refresh=False, legacy=False, **kwargs):
    """
    The Large model trained on Libri-Light 60k hours + CommonVoice + Switchboard + Fisher
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs[
        "ckpt"
    ] = "https://dl.fbaipublicfiles.com/fairseq/wav2vec/w2v_large_lv_fsh_swbd_cv.pt"
    if not legacy:
        kwargs[
            "ckpt"
        ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/w2v_large_lv_fsh_swbd_cv.pt"
    return wav2vec2_custom(refresh=refresh, legacy=legacy, **kwargs)


def xlsr_53(refresh=False, legacy=False, **kwargs):
    """
    The wav2vec 2.0 model trained on multilingual presented in https://arxiv.org/abs/2006.13979
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs["ckpt"] = "https://dl.fbaipublicfiles.com/fairseq/wav2vec/xlsr_53_56k.pt"
    if not legacy:
        kwargs[
            "ckpt"
        ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/xlsr_53_56k.pt"
    return wav2vec2_custom(refresh=refresh, legacy=legacy, **kwargs)


def xls_r_300m(refresh=False, legacy=False, **kwargs):
    """
    XLS-R, this smallest size has the same parameters as the Largs model of wav2vec 2.0 and HuBERT
    """
    kwargs["ckpt"] = "https://dl.fbaipublicfiles.com/fairseq/wav2vec/xlsr2_300m.pt"
    if not legacy:
        kwargs[
            "ckpt"
        ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/xlsr2_300m.pt"
    return wav2vec2_custom(refresh=refresh, legacy=legacy, **kwargs)


def xls_r_1b(refresh=False, legacy=False, **kwargs):
    kwargs[
        "ckpt"
    ] = "https://dl.fbaipublicfiles.com/fairseq/wav2vec/xlsr2_960m_1000k.pt"
    if not legacy:
        kwargs[
            "ckpt"
        ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/xlsr2_960m_1000k.pt"
    return wav2vec2_custom(refresh=refresh, legacy=legacy, **kwargs)


def xls_r_2b(refresh=False, legacy=False, **kwargs):
    kwargs["ckpt"] = "https://dl.fbaipublicfiles.com/fairseq/wav2vec/xlsr2_2B_1000k.pt"
    if not legacy:
        kwargs[
            "ckpt"
        ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/xlsr2_2B_1000k.pt"
    return wav2vec2_custom(refresh=refresh, legacy=legacy, **kwargs)


def wav2vec2_conformer_relpos(refresh=False, legacy=False, **kwargs):
    kwargs[
        "ckpt"
    ] = "https://dl.fbaipublicfiles.com/fairseq/conformer/wav2vec2/librilight/LL_relpos_PT_no_FT"
    if not legacy:
        kwargs[
            "ckpt"
        ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/LL_relpos_PT_no_FT.pt"
    return wav2vec2_custom(refresh=refresh, legacy=legacy, **kwargs)


def wav2vec2_conformer_rope(refresh=False, legacy=False, **kwargs):
    kwargs[
        "ckpt"
    ] = "https://dl.fbaipublicfiles.com/fairseq/conformer/wav2vec2/librilight/LL_rope_PT_no_FT"
    if not legacy:
        kwargs[
            "ckpt"
        ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/LL_rope_PT_no_FT.pt"
    return wav2vec2_custom(refresh=refresh, legacy=legacy, **kwargs)


def wav2vec2_large_voxpopuli_100k(refresh=False, legacy=False, **kwargs):
    kwargs[
        "ckpt"
    ] = "https://dl.fbaipublicfiles.com/voxpopuli/models/wav2vec2_large_100k.pt"
    if not legacy:
        kwargs[
            "ckpt"
        ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/wav2vec2_large_100k.pt"
    return wav2vec2_custom(refresh=refresh, legacy=legacy, **kwargs)


def wav2vec2_base_s2st_es_voxpopuli(refresh=False, legacy=False, **kwargs):
    kwargs[
        "ckpt"
    ] = "https://dl.fbaipublicfiles.com/fairseq/speech_to_speech/s2st_finetuning/w2v2/es/transformer_B.pt"
    if not legacy:
        kwargs[
            "ckpt"
        ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/wav2vec2_base_s2st_es_voxpopuli.pt"
    return wav2vec2_custom(refresh=refresh, legacy=legacy, **kwargs)


# FIXME: the official fairseq checkpoint link is down currently
# def wav2vec2_large_s2st_es_voxpopuli(refresh=False, legacy=False, **kwargs):
#     kwargs[
#         "ckpt"
#     ] = "https://dl.fbaipublicfiles.com/fairseq/speech_to_speech/s2st_finetuning/w2v2/es/transformer_L.pt"
#     if not legacy:
#         kwargs[
#             "ckpt"
#         ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/wav2vec2-large-s2st-es-voxpopuli.pt"
#     return wav2vec2_custom(refresh=refresh, legacy=legacy, **kwargs)


def wav2vec2_conformer_large_s2st_es_voxpopuli(refresh=False, legacy=False, **kwargs):
    kwargs[
        "ckpt"
    ] = "https://dl.fbaipublicfiles.com/fairseq/speech_to_speech/s2st_finetuning/w2v2/es/conformer_L.pt"
    if not legacy:
        kwargs[
            "ckpt"
        ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/wav2vec2_conformer_large_s2st_es_voxpopuli.pt"
    return wav2vec2_custom(refresh=refresh, legacy=legacy, **kwargs)


def wav2vec2_base_s2st_en_librilight(refresh=False, legacy=False, **kwargs):
    kwargs[
        "ckpt"
    ] = "https://dl.fbaipublicfiles.com/fairseq/speech_to_speech/s2st_finetuning/w2v2/en/transformer_B.pt"
    if not legacy:
        kwargs[
            "ckpt"
        ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/wav2vec2_base_s2st_en_librilight.pt"
    return wav2vec2_custom(refresh=refresh, legacy=legacy, **kwargs)


def wav2vec2_conformer_large_s2st_en_librilight(refresh=False, legacy=False, **kwargs):
    kwargs[
        "ckpt"
    ] = "https://dl.fbaipublicfiles.com/fairseq/speech_to_speech/s2st_finetuning/w2v2/en/conformer_L.pt"
    if not legacy:
        kwargs[
            "ckpt"
        ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/wav2vec2_conformer_large_s2st_en_librilight.pt"
    return wav2vec2_custom(refresh=refresh, legacy=legacy, **kwargs)



"""
Model interfaces

Authors:
  * Leo 2022
"""

from typing import List, Tuple

import torch
import torch.nn as nn

__all__ = [
    "AbsUpstream",
    "AbsFeaturizer",
    "AbsFrameModel",
    "AbsUtteranceModel",
]


class AbsUpstream(nn.Module):
    """
    The upstream model should follow this interface. Please subclass it.
    """

    @property
    def num_layer(self) -> int:
        """
        number of hidden states
        """
        raise NotImplementedError

    @property
    def hidden_sizes(self) -> List[int]:
        """
        hidden size of each hidden state
        """
        raise NotImplementedError

    @property
    def downsample_rates(self) -> List[int]:
        """
        downsample rate from 16 KHz waveforms for each hidden state
        """
        raise NotImplementedError

    def forward(
        self, wavs: torch.FloatTensor, wavs_len: torch.LongTensor
    ) -> Tuple[List[torch.FloatTensor], List[torch.LongTensor]]:
        """
        Args:
            wavs (torch.FloatTensor): (batch_size, seq_len, 1)
            wavs_len (torch.LongTensor): (batch_size, )

        Returns:
            tuple:

            1. all_hs (List[torch.FloatTensor]): all the hidden states
            2. all_hs_len (List[torch.LongTensor]): the lengths for all the hidden states
        """
        raise NotImplementedError


class AbsFeaturizer(nn.Module):
    """
    The featurizer should follow this interface. Please subclass it.
    The featurizer's mission is to reduce (standardize) the multiple hidden
    states from :obj:`AbsUpstream` into a single hidden state, so that
    the downstream model can use it as a conventional representation.
    """

    @property
    def output_size(self) -> int:
        """
        The output size after hidden states reduction
        """
        raise NotImplementedError

    @property
    def downsample_rate(self) -> int:
        """
        The downsample rate from 16 KHz waveform of the reduced single hidden state
        """
        raise NotImplementedError

    def forward(
        self, all_hs: List[torch.FloatTensor], all_hs_len: List[torch.LongTensor]
    ) -> Tuple[torch.FloatTensor, torch.LongTensor]:
        """
        Args:
            all_hs (List[torch.FloatTensor]): all the hidden states
            all_hs_len (List[torch.LongTensor]): the lengths for all the hidden states

        Returns:
            tuple:

            1. hs (torch.FloatTensor)
            2. hs_len (torch.LongTensor)
        """
        raise NotImplementedError


class AbsFrameModel(nn.Module):
    """
    The frame-level model interface.
    """

    @property
    def input_size(self) -> int:
        raise NotImplementedError

    @property
    def output_size(self) -> int:
        raise NotImplementedError

    def forward(
        self, x: torch.FloatTensor, x_len: torch.LongTensor
    ) -> Tuple[torch.FloatTensor, torch.LongTensor]:
        raise NotImplementedError


class AbsUtteranceModel(nn.Module):
    """
    The utterance-level model interface, which pools the temporal dimension.
    """

    @property
    def input_size(self) -> int:
        raise NotImplementedError

    @property
    def output_size(self) -> int:
        raise NotImplementedError

    def forward(
        self, x: torch.FloatTensor, x_len: torch.LongTensor
    ) -> torch.FloatTensor:
        raise NotImplementedError



"""
The probing model following Hear Benchmark

Authors:
  * Hear Team 2021
  * Leo 2022
"""


from typing import List

import torch

import s3prl.nn.pooling as pooling

__all__ = ["HearFullyConnectedPrediction"]


class HearFullyConnectedPrediction(torch.nn.Module):
    """
    The specific prediction head used in the Hear Benchmark.
    Modified from: https://github.com/hearbenchmark/hear-eval-kit/blob/855964977238e89dfc76394aa11c37010edb6f20/heareval/predictions/task_predictions.py#L142

    Args:
        input_size (int): input_size
        output_size (int): output_size
        hidden_size (int): hidden size across all layers. Default: 1024
        hidden_layers (int): number of hidden layers, all in :code:`hidden_size`. Default: 2
        norm_after_activation (bool): whether to norm after activation. Default: False
        dropout (float): dropout ratio. Default: 0.1
        initialization (str): initialization method name available in :obj:`torch.nn.init`
        hidden_norm (str): normalization method name available in :obj:`torch.nn`
        pooling_type (str): the pooling class name in :obj:`s3prl.nn.pooling`. Default: MeanPooling
        pooling_conf (dict): the arguments for initializing the pooling class.
            Default: empty dict
    """

    def __init__(
        self,
        input_size: int,
        output_size: int,
        hidden_size: int = 1024,
        hidden_layers: int = 2,
        norm_after_activation: bool = False,
        dropout: float = 0.1,
        initialization: str = "xavier_uniform_",
        hidden_norm: str = "BatchNorm1d",
        pooling_type: str = None,
        pooling_conf: dict = None,
    ):
        super().__init__()
        self._input_size = input_size
        self._output_size = output_size
        initialization = getattr(torch.nn.init, initialization)
        hidden_norm = getattr(torch.nn, hidden_norm)

        curdim = input_size
        if pooling_type is not None:
            pooling_cls = getattr(pooling, pooling_type)
            self.pooling = pooling_cls(input_size, **(pooling_conf or {}))
            curdim = self.pooling.output_size

        hidden_modules: List[torch.nn.Module] = []
        last_activation = "linear"
        if hidden_layers:
            for i in range(hidden_layers):
                linear = torch.nn.Linear(curdim, hidden_size)
                initialization(
                    linear.weight,
                    gain=torch.nn.init.calculate_gain(last_activation),
                )
                hidden_modules.append(linear)
                if not norm_after_activation:
                    hidden_modules.append(hidden_norm(hidden_size))
                hidden_modules.append(torch.nn.Dropout(dropout))
                hidden_modules.append(torch.nn.ReLU())
                if norm_after_activation:
                    hidden_modules.append(hidden_norm(hidden_size))
                curdim = hidden_size
                last_activation = "relu"

            self.hidden = torch.nn.Sequential(*hidden_modules)
        else:
            self.hidden = torch.nn.Identity()  # type: ignore
        self.projection = torch.nn.Linear(curdim, output_size)

        initialization(
            self.projection.weight, gain=torch.nn.init.calculate_gain(last_activation)
        )

    @property
    def input_size(self) -> int:
        return self._input_size

    @property
    def output_size(self) -> int:
        return self._output_size

    def forward(self, x, x_len) -> torch.Tensor:
        """
        Args:
            x (torch.FloatTensor): (batch_size, seq_len, input_size)
            x_len (torch.LongTensor): (batch_size, )

        Returns:
            tuple:

            1. y (torch.FloatTensor)
            2. y_len (torch.LongTensor)

            if :code:`pooling_type` is None, :code:`y` is (batch_size, seq_len, output_size) and :code:`y_len` is (batch_size, )
            if not None, :code:`y` is (batch_size, output_size) and :code:`y_len` is (batch_size, ) in all 1s.
        """
        if hasattr(self, "pooling"):
            x = self.pooling(x, x_len)
            x_len = x.new_ones(len(x))

        shape = x.shape
        if len(shape) == 3:
            bs, ts, hidden_size = x.shape
            x = x.reshape(bs * ts, hidden_size)

        x = self.hidden(x)
        x = self.projection(x)

        if len(shape) == 3:
            x = x.reshape(bs, ts, -1)

        return x, x_len



from pathlib import Path

import pandas as pd
import torch
import torchaudio

from .superb_sid import SuperbSID

torchaudio.set_audio_backend("sox_io")


class CommonExample(SuperbSID):
    def default_config(self) -> dict:
        config = super().default_config()
        config["prepare_data"] = {}
        config["train"] = dict(
            total_steps=10,
            log_step=1,
            eval_step=5,
            save_step=5,
            gradient_clipping=1.0,
            gradient_accumulate=1,
            valid_metric="accuracy",
            valid_higher_better=True,
            auto_resume=True,
        )
        return config

    def prepare_data(
        self,
        prepare_data: dict,
        target_dir: str,
        cache_dir: str,
        get_path_only: bool = False,
    ):
        target_dir: Path = Path(target_dir)

        wavs = [torch.randn(1, 16000 * 2) for i in range(5)]
        wav_paths = []
        for idx, wav in enumerate(wavs):
            wav_path = str(Path(target_dir) / f"{idx}.wav")
            torchaudio.save(wav_path, wav, sample_rate=16000)
            wav_paths.append(wav_path)

        ids = [Path(path).stem for path in wav_paths]
        labels = ["a", "a", "b", "c", "d"]

        df = pd.DataFrame({"id": ids, "wav_path": wav_paths, "label": labels})
        train_csv, valid_csv, test_csv = (
            target_dir / "train.csv",
            target_dir / "valid.csv",
            target_dir / "test.csv",
        )
        df.iloc[:3].to_csv(train_csv)
        df.iloc[3:4].to_csv(valid_csv)
        df.iloc[4:].to_csv(test_csv)

        return train_csv, valid_csv, [test_csv]



"""
Parse VoxCeleb1 corpus for verification

Authors:
  * Haibin Wu 2022
"""

import logging
from pathlib import Path

from tqdm import tqdm

from s3prl.util.download import _download

from .base import Corpus

SPLIT_FILE_URL = "https://www.robots.ox.ac.uk/~vgg/data/voxceleb/meta/iden_split.txt"
TRIAL_FILE_URL = "https://openslr.magicdatatech.com/resources/49/voxceleb1_test_v2.txt"

__all__ = [
    "VoxCeleb1SV",
]


class VoxCeleb1SV(Corpus):
    def __init__(
        self, dataset_root: str, download_dir: str, force_download: bool = True
    ) -> None:
        self.dataset_root = Path(dataset_root).resolve()

        train_path, valid_path, test_path, speakerid2label = self.format_path(
            self.dataset_root, download_dir, force_download
        )
        self.categories = speakerid2label
        self.train_data = self.path2data(train_path, speakerid2label)
        self.valid_data = self.path2data(valid_path, speakerid2label)
        self.test_data = {
            self.path2uid(path): {"wav_path": path, "label": None} for path in test_path
        }
        self.test_trials = self.format_test_trials(download_dir, force_download)

    @classmethod
    def path2uid(cls, path):
        return "-".join(Path(path).parts[-3:])

    @classmethod
    def path2data(cls, paths, speakerid2label):
        data = {
            cls.path2uid(path): {
                "wav_path": path,
                "label": speakerid2label[Path(path).parts[-3]],
            }
            for path in paths
        }
        return data

    @staticmethod
    def format_path(dataset_root, download_dir, force_download: bool):
        split_filename = SPLIT_FILE_URL.split("/")[-1]
        split_filepath = Path(download_dir) / split_filename
        _download(split_filepath, SPLIT_FILE_URL, refresh=force_download)

        usage_list = open(split_filepath, "r").readlines()
        train, valid, test = [], [], []
        test_list = [
            item
            for item in usage_list
            if int(item.split(" ")[1].split("/")[0][2:]) in range(10270, 10310)
        ]
        usage_list = list(set(usage_list).difference(set(test_list)))
        test_list = [item.split(" ")[1] for item in test_list]

        logging.info("search specified wav name for each split")
        speakerids = []

        for string in tqdm(usage_list, desc="Search train, dev wavs"):
            pair = string.split()
            index = pair[0]
            x = list(dataset_root.glob("dev/wav/" + pair[1]))
            speakerStr = pair[1].split("/")[0]
            if speakerStr not in speakerids:
                speakerids.append(speakerStr)
            if int(index) == 1 or int(index) == 3:
                train.append(str(x[0]))
            elif int(index) == 2:
                valid.append(str(x[0]))
            else:
                raise ValueError

        speakerids = sorted(speakerids)
        speakerid2label = {}
        for idx, spk in enumerate(speakerids):
            speakerid2label[spk] = idx

        for string in tqdm(test_list, desc="Search test wavs"):
            x = list(dataset_root.glob("test/wav/" + string.strip()))
            test.append(str(x[0]))
        logging.info(
            f"finish searching wav: train {len(train)}; valid {len(valid)}; test {len(test)} files found"
        )

        return train, valid, test, speakerid2label

    @classmethod
    def format_test_trials(cls, download_dir: str, force_download: bool):
        trial_filename = TRIAL_FILE_URL.split("/")[-1]
        trial_filepath = Path(download_dir) / trial_filename
        _download(trial_filepath, TRIAL_FILE_URL, refresh=force_download)

        trial_list = open(trial_filepath, "r").readlines()
        test_trials = []
        for string in tqdm(trial_list, desc="Prepare testing trials"):
            pair = string.split()
            test_trials.append(
                (int(pair[0]), cls.path2uid(pair[1]), cls.path2uid(pair[2]))
            )

        return test_trials

    @property
    def all_data(self):
        return self.train_data, self.valid_data, self.test_data, self.test_trials

    @property
    def data_split_ids(self):
        return None



import copy

import torch
import torch.nn as nn
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

from s3prl import Output
from s3prl.nn.vq_apc import VqApcLayer

__all__ = [
    "RnnApc",
]


class RnnApc(nn.Module):
    """
    The RNN model.
    Currently supporting upstreams models of APC, VQ-APC.
    """

    def __init__(self, input_size, hidden_size, num_layers, dropout, residual, vq=None):
        """
        Args:
            input_size (int):
                An int indicating the input feature size, e.g., 80 for Mel.
            hidden_size (int):
                An int indicating the RNN hidden size.
            num_layers (int):
                An int indicating the number of RNN layers.
            dropout (float):
                A float indicating the RNN dropout rate.
            residual (bool):
                A bool indicating whether to apply residual connections.
        """
        super(RnnApc, self).__init__()

        assert num_layers > 0
        self.hidden_size = hidden_size
        self.code_dim = hidden_size  # ToDo: different size?
        self.num_layers = num_layers
        in_sizes = [input_size] + [hidden_size] * (num_layers - 1)
        out_sizes = [hidden_size] * num_layers
        self.rnn_layers = nn.ModuleList(
            [
                nn.GRU(input_size=in_size, hidden_size=out_size, batch_first=True)
                for (in_size, out_size) in zip(in_sizes, out_sizes)
            ]
        )

        self.rnn_dropout = nn.Dropout(dropout)

        self.rnn_residual = residual

        #  Create N-group VQ layers (Last layer only)
        self.apply_vq = vq is not None
        if self.apply_vq:
            self.vq_layers = []
            vq_config = copy.deepcopy(vq)
            codebook_size = vq_config.pop("codebook_size")
            self.vq_code_dims = vq_config.pop("code_dim")
            assert len(self.vq_code_dims) == len(codebook_size)
            assert sum(self.vq_code_dims) == hidden_size
            for cs, cd in zip(codebook_size, self.vq_code_dims):
                self.vq_layers.append(
                    VqApcLayer(
                        input_size=cd, code_dim=cd, codebook_size=cs, **vq_config
                    )
                )
            self.vq_layers = nn.ModuleList(self.vq_layers)

        # TODO: Start with a high temperature and anneal to a small one.
        # Final regression layer
        self.postnet = nn.Linear(hidden_size, input_size)

    def create_msg(self):
        msg_list = []
        msg_list.append(
            "Model spec.| Method = APC\t| Apply VQ = {}\t".format(self.apply_vq)
        )
        msg_list.append(
            "           | n layers = {}\t| Hidden dim = {}".format(
                self.num_layers, self.hidden_size
            )
        )
        return msg_list

    def report_ppx(self):
        if self.apply_vq:
            # ToDo: support more than 2 groups
            ppx = [m.report_ppx() for m in self.vq_layers] + [None]
            return ppx[0], ppx[1]
        else:
            return None, None

    def report_usg(self):
        if self.apply_vq:
            # ToDo: support more than 2 groups
            usg = [m.report_usg() for m in self.vq_layers] + [None]
            return usg[0], usg[1]
        else:
            return None, None

    def forward(self, frames_BxLxM, seq_lengths_B, testing=False):
        """
        Args:
            frames_BxLxM (torch.LongTensor):
                A 3d-tensor representing the input features.
            seq_lengths_B (list):
                A list containing the sequence lengths of `frames_BxLxM`.
            testing (bool):
                A bool indicating training or testing phase.
                Default: False
        Return:
            Output (s3prl.Output):
                An Output module that contains `hidden_states` and `prediction`

                hidden_states (hiddens_NxBxLxH):
                    The RNN hidden representations across all layers.
                prediction (predicted_BxLxM):
                    The predicted output; used for training.
        """

        max_seq_len = frames_BxLxM.size(1)

        # N is the number of RNN layers.
        hiddens_NxBxLxH = []

        # RNN
        # Prepare initial packed RNN input.
        packed_rnn_inputs = pack_padded_sequence(
            frames_BxLxM, seq_lengths_B, batch_first=True, enforce_sorted=False
        )
        for i, rnn_layer in enumerate(self.rnn_layers):
            # https://discuss.pytorch.org/t/rnn-module-weights-are-not-part-of-single-contiguous-chunk-of-memory/6011/14
            rnn_layer.flatten_parameters()
            packed_rnn_outputs, _ = rnn_layer(packed_rnn_inputs)

            # Unpack RNN output of current layer.
            rnn_outputs_BxLxH, _ = pad_packed_sequence(
                packed_rnn_outputs, batch_first=True, total_length=max_seq_len
            )
            # Apply dropout to output.
            rnn_outputs_BxLxH = self.rnn_dropout(rnn_outputs_BxLxH)

            # Apply residual connections.
            if self.rnn_residual and i > 0:
                # Unpack the original input.
                rnn_inputs_BxLxH, _ = pad_packed_sequence(
                    packed_rnn_inputs, batch_first=True, total_length=max_seq_len
                )
                rnn_outputs_BxLxH += rnn_inputs_BxLxH

            hiddens_NxBxLxH.append(rnn_outputs_BxLxH)

            # VQ at last layer only
            if self.apply_vq and (i == len(self.rnn_layers) - 1):
                q_feat = []
                offet = 0
                for vq_layer, cd in zip(self.vq_layers, self.vq_code_dims):
                    q_f = vq_layer(
                        rnn_outputs_BxLxH[:, :, offet : offet + cd], testing
                    ).output
                    q_feat.append(q_f)
                    offet += cd
                rnn_outputs_BxLxH = torch.cat(q_feat, dim=-1)

            # Prepare packed input for the next layer.
            # Note : enforce sorted = False might lead to CUDNN_STATUS_EXECUTION_FAILED
            if i < len(self.rnn_layers) - 1:
                packed_rnn_inputs = pack_padded_sequence(
                    rnn_outputs_BxLxH,
                    seq_lengths_B,
                    batch_first=True,
                    enforce_sorted=False,
                )
        # Only return last layer feature
        feature = hiddens_NxBxLxH[-1]

        # Generate final output from codes.
        predicted_BxLxM = self.postnet(rnn_outputs_BxLxH)
        return Output(hidden_states=feature, prediction=predicted_BxLxM)



"""
Speech-to-text based recipes
"""

from .superb_asr import SuperbASR
from .superb_pr import SuperbPR
from .superb_sf import SuperbSF

__all__ = [
    "SuperbASR",
    "SuperbPR",
    "SuperbSF",
]



"""
Simple categorical encoder

Authors:
  Leo 2022
"""

from typing import List


class CategoryEncoder:
    def __init__(self, category: List[str]) -> None:
        self.category = list(sorted(set(category)))

    def __len__(self) -> int:
        return len(self.category)

    def encode(self, label: str) -> int:
        return self.category.index(label)

    def decode(self, index: int) -> str:
        return self.category[index]


class CategoryEncoders:
    def __init__(self, categories: List[List[str]]) -> None:
        self.categories = [CategoryEncoder(c) for c in categories]

    def __len__(self) -> int:
        return sum([len(c) for c in self.categories])

    def __iter__(self):
        for category in self.categories:
            yield category

    def encode(self, labels: List[str]) -> List[int]:
        assert len(labels) == len(self.categories)
        return [
            encoder.encode(label) for label, encoder in zip(labels, self.categories)
        ]

    def decode(self, indices: List[int]) -> List[str]:
        return [
            encoder.decode(index) for index, encoder in zip(indices, self.categories)
        ]



# -*- coding: utf-8 -*-
# @Time    : 8/25/21 5:25 PM
# @Author  : Yuan Gong
# @Affiliation  : Massachusetts Institute of Technology
# @Email   : yuangong@mit.edu
# @File    : hubconf.py

# Authors
# - Leo

from s3prl.util.download import _urls_to_filepaths

from .expert import UpstreamExpert as _UpstreamExpert


# Frame-based SSAST
# 1s for speech commands, 6s for IEMOCAP, 10s for SID
def ssast_frame_base(refresh: bool = False, window_secs: float = 1.0, **kwargs):
    ckpt = _urls_to_filepaths(
        "https://www.dropbox.com/s/nx6nl4d4bl71sm8/SSAST-Base-Frame-400.pth?dl=1",
        refresh=refresh,
    )
    return _UpstreamExpert(ckpt, "base_f", window_secs)


# Patch-based SSAST
# 1s for speech commands, 6s for IEMOCAP, 10s for SID
def ssast_patch_base(refresh: bool = False, window_secs: float = 1.0, **kwargs):
    ckpt = _urls_to_filepaths(
        "https://www.dropbox.com/s/ewrzpco95n9jdz6/SSAST-Base-Patch-400.pth?dl=1",
        refresh=refresh,
    )
    return _UpstreamExpert(ckpt, "base_p", window_secs)



from torch import nn

from s3prl import Output
from s3prl.nn.transformer_mockingjay import (
    ACT2FN,
    TransformerConfig,
    TransformerLayerNorm,
)


class PredictorMockingjay(nn.Module):
    """
    The predictor model for SSL pre-training tasks.
    Currently supporting SSL problems of Mockingjay, Tera, and Audio Albert.
    """

    def __init__(self, config, output_dim, input_dim=None, **kwargs):
        """
        Args:
            config (TransformerConfig):
                A `TransformerConfig` class instance with the configuration to build a new model,
                can also be a `dict` that initializes the TransformerConfig class
            output_dim (int):
                The output dimension of predictor
            input_dim (int):
                The input dimension of predictor, if `None` is given, then use the `hidden_size` defined in `config`.
                Default: None
        """

        super(PredictorMockingjay, self).__init__()
        if type(config) is dict:
            config = TransformerConfig(**config)
        self.output_size = output_dim
        if input_dim is None:
            self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        else:
            self.dense = nn.Linear(input_dim, config.hidden_size)
        if isinstance(config.hidden_act, str):
            self.transform_act_fn = ACT2FN[config.hidden_act]
        else:
            self.transform_act_fn = config.hidden_act
        self.LayerNorm = TransformerLayerNorm(
            config.hidden_size, eps=config.layer_norm_eps
        )
        self.output = nn.Linear(config.hidden_size, self.output_size)

    def forward(self, inputs, output_states=False):
        """
        Args:
            inputs (torch.LongTensor):
                A torch.LongTensor of shape [batch_size, sequence_length, input_dim]
            output_states (bool):
                A boolean which controls whether to return the `hidden_states` of the predictor.
                Default: False
        Return:
            Output (s3prl.Output):
                An Output module that contains `prediction` and/or `hidden_states`.
        """
        hidden_states = inputs.hidden_states
        hidden_states = self.dense(hidden_states)
        hidden_states = self.transform_act_fn(hidden_states)
        hidden_states = self.LayerNorm(hidden_states)
        prediction = self.output(hidden_states)
        if output_states:
            return Output(hidden_states=hidden_states, prediction=prediction)
        else:
            return Output(prediction=prediction)



"""BYOL for Audio: Common definitions and utilities."""
import datetime
import logging
import os
import random
from argparse import Namespace
from pathlib import Path

import yaml

try:
    import pickle5 as pickle
except ImportError:
    import pickle

import numpy as np
import torch
import torchaudio

torchaudio.set_audio_backend("sox_io")


def get_timestamp():
    """ex) Outputs 202104220830"""
    return datetime.datetime.now().strftime("%y%m%d%H%M")


def load_yaml_config(path_to_config):
    """Loads yaml configuration settings as an EasyDict object."""
    path_to_config = Path(path_to_config)
    assert path_to_config.is_file()
    with open(path_to_config) as f:
        yaml_contents = yaml.safe_load(f)
    cfg = Namespace(**yaml_contents)
    return cfg


def get_logger(name):
    logging.basicConfig(
        format="%(asctime)s %(levelname)s %(message)s",
        datefmt="%Y-%m-%d %H:%M",
        level=logging.DEBUG,
    )
    logger = logging.getLogger(name)
    return logger



import random
from typing import List, Tuple

import librosa
import torch
import torchaudio

from . import Dataset

torchaudio.set_audio_backend("sox_io")


class LoadAudio(Dataset):
    """
    Args:
        start_secs: use None if load from start
        end_secs: use None if load to end
    """

    def __init__(
        self,
        filepaths: List[str],
        start_secs: List[float] = None,
        end_secs: List[float] = None,
        sox_effects: Tuple[Tuple[str]] = None,
        individual_sox_effects: List[Tuple[Tuple[str]]] = None,
        max_secs: float = None,
        generator: random.Random = None,
        sample_rate: int = 16000,
    ) -> None:
        super().__init__()
        self.filepaths = filepaths
        self.start_secs = start_secs
        self.end_secs = end_secs
        if generator is None:
            generator = random.Random(12345678)
        self.generator = generator

        self.sample_rate = sample_rate
        self.max_secs = max_secs

        assert int(start_secs is not None) + int(end_secs is not None) in [
            0,
            2,
        ], "start_secs and end_secs must both be given if anyone is given"

        assert (
            int(sox_effects is not None) + int(individual_sox_effects is not None) <= 1
        )
        if sox_effects is not None:
            individual_sox_effects = [sox_effects for _ in range(len(filepaths))]
        self.individual_sox_effects = individual_sox_effects

    def __len__(self):
        return len(self.filepaths)

    def __getitem__(self, index: int):
        start_sec = None if self.start_secs is None else self.start_secs[index]
        start_sec = start_sec or 0.0
        end_sec = None if self.end_secs is None else self.end_secs[index]
        duration = None if end_sec is None else (self.end_secs[index] - start_sec)

        y, sr = librosa.load(
            self.filepaths[index],
            sr=self.sample_rate,
            offset=start_sec,
            duration=duration,
        )
        assert sr == self.sample_rate
        wav = torch.FloatTensor(y).view(1, -1)

        if self.individual_sox_effects is not None:
            wav, sr = torchaudio.sox_effects.apply_effects_tensor(
                wav, sr, effects=self.individual_sox_effects[index]
            )

        if sr != self.sample_rate:
            wav, sr = torchaudio.transforms.Resample(sr, self.sample_rate)(wav)

        if self.max_secs is not None:
            secs = wav.size(-1) / self.sample_rate
            if secs > self.max_secs:
                max_samples = round(self.max_secs * self.sample_rate)
                start = self.generator.randint(0, wav.size(-1) - max_samples)
                wav = wav[:, start : start + max_samples]

        wav = wav.view(-1)
        return {
            "wav_path": self.filepaths[index],
            "wav_len": len(wav),
            "wav": wav,
        }
